"""
AUTHOR : Michael Cretignier 
EDITOR : Khaled Al Moulla

Collection of classes.
"""

###
### MODULES


import matplotlib.pyplot             as     plt
import numpy                         as     np
import pandas                        as     pd

from   matplotlib.collections        import LineCollection
from   scipy.interpolate             import interp1d
from   statsmodels.stats.weightstats import DescrStatsW

from ..ANTARESS_conversions.KitCat import Kitcat_functions as myf
from ..ANTARESS_conversions.KitCat.Kitcat_functions import rm_outliers as rm_out


#np.warnings.filterwarnings('ignore', category=RuntimeWarning)

###
### CLASSES

class tableXY(object):

    def __init__(self, x, y, *yerr):
        self.stats = pd.DataFrame({},index=[0])
        self.y = np.array(y)  #vector of y

        if x is None:# for a fast table initialisation
            x = np.arange(len(y))
        self.x = np.array(x)  #vector of x
        
        try:
            np.sum(self.y) 
        except: #in case of None
            self.y = np.zeros(len(self.x))
            yerr = [np.ones(len(self.y))]
                
        if len(x)!=len(y):
            print('X et Y have no the same lenght')

        if len(yerr)!=0:
            if len(yerr)==1:
                self.yerr = np.array(yerr[0])
                self.xerr =  np.zeros(len(self.x))
            elif len(yerr)==2:
                self.xerr = np.array(yerr[0])
                self.yerr = np.array(yerr[1])
        else :
            if sum(~np.isnan(self.y.astype('float'))):
                self.yerr = np.ones(len(self.x))*myf.mad(rm_out(self.y.astype('float'),m=2,kind='sigma')[1])
                if not np.sum(abs(self.yerr)):
                    self.yerr = np.ones(len(self.x))
            else:
                self.yerr = np.ones(len(self.x))
            self.xerr =  np.zeros(len(self.x))

    def clip(self, min=[None,None], max=[None,None], replace=True, invers=False):
        """This function seems sometimes to not work without any reason WARNING"""
        min2 = np.array(min).copy() ; max2 = np.array(max).copy()
        if min2[1] == None:
            min2[1] = np.nanmin(self.y)-1
        if max2[1] == None:
            max2[1] = np.nanmax(self.y)+1
        masky = (self.y<=max2[1])&(self.y>=min2[1])
        if min2[0] == None:
            min2[0] = np.nanmin(self.x)-1
        if max2[0] == None:
            max2[0] = np.nanmax(self.x)+1
        maskx = (self.x<=max2[0])&(self.x>=min2[0])
        mask = maskx&masky
        try:
            self.clip_mask = self.clip_mask&mask
        except:    
            self.clip_mask = mask

        if invers:
            mask = ~mask
        self.clipped = tableXY(self.x[mask],self.y[mask],self.xerr[mask],self.yerr[mask])
        if replace==True:
            self.x = self.x[mask] ; self.y = self.y[mask] ; self.yerr = self.yerr[mask] ; self.xerr= self.xerr[mask]
        else:
            self.clipx = self.x[mask] ; self.clipy=self.y[mask] ; self.clipyerr = self.yerr[mask] ; self.clipxerr=self.xerr[mask]

    def copy(self):
        return tableXY(self.x.copy(),self.y.copy(),self.xerr.copy(),self.yerr.copy())

    def fit_line(self, perm=1000, Draw=False, color='k', info=False, fontsize=13, label=True, compute_r=True, offset=True, recenter=True, info_printed=['r','s','i','rms']):
        k = perm
        self.yerr[self.yerr==0] = [np.min(self.yerr),0.1][np.min(self.yerr)==0] #to avoid 0 value
        
        w = 1/self.yerr**2    
        if offset:    
            A = np.array([(self.x-np.mean(self.x)*int(recenter)),np.ones(len(self.x))]).T
        else:
            A = np.array([self.x]).T

        A = A *np.sqrt(w)[:,np.newaxis]
        B = np.array([self.y]*(k+1)).T
        noise = np.random.randn(np.shape(B)[0],np.shape(B)[1])/np.sqrt(w)[:,np.newaxis] ; noise[:,0] = 0
        B = B + noise
        Bmean = np.sum(B*w[:,np.newaxis],axis=0)/np.sum(w)*int(recenter)
        Brms = np.sqrt(np.sum(((B-Bmean)**2*w[:,np.newaxis]),axis=0)/np.sum(w))
        B = B*np.sqrt(w)[:,np.newaxis]
        Cmean = np.sum(self.x*w,axis=0)/np.sum(w)*int(recenter)
        Crms = np.sqrt(np.sum(((self.x-Cmean)**2*w),axis=0)/np.sum(w))

        self.s = np.linalg.lstsq(A,B,rcond=None)[0][0]
        if offset:
            self.i = np.linalg.lstsq(A,B,rcond=None)[0][1]      
        else:
            self.i = self.s*0

        self.lin_slope_w = np.mean(self.s)
        self.lin_errslope_w = np.std(self.s)
        
        self.lin_intercept_w = np.mean(self.i)
        self.lin_errintercept_w = np.std(self.i)

        self.stats['lin_slope_w'] = self.lin_slope_w
        self.stats['lin_slope_w_std'] = self.lin_errslope_w
        self.stats['lin_intercept_w'] = self.lin_intercept_w
        self.stats['lin_intercept_w_std'] = self.lin_errintercept_w     

        if compute_r:
            self.r = self.s*Crms/Brms        
            self.r_pearson_w = np.mean(self.r)
            self.r_errpearson_w = np.std(self.r)
        else:
            self.r_pearson_w = np.inf
            self.r_errpearson_w = np.inf
            
        self.stats['r_pearson_w'] = self.r_pearson_w
        self.stats['r_pearson_w_std'] = self.r_errpearson_w
        
        
        temp = tableXY(self.x, self.y-((self.x-np.mean(self.x)*int(recenter))*self.lin_slope_w+self.lin_intercept_w), self.yerr)
        temp.rms_w()
        self.fit_line_rms = temp.rms
        
        if Draw:
            if label:
                lbl = ''
                if perm==1:
                    symb = '|'
                    lbl = ' $\\mathcal{R}$=%.2f %s S=%.2f %s I=%.2f %s rms=%.2f'%(self.r_pearson_w,symb,self.lin_slope_w,symb,self.lin_intercept_w,symb,temp.rms)
                else:
                    symb = '\n'
                    txt = {
                        'r':['$\\mathcal{R}$',self.r_pearson_w, self.r_errpearson_w],
                        's':['S',self.lin_slope_w, self.lin_errslope_w],
                        'i':['I',self.lin_intercept_w, self.lin_errintercept_w],
                        'rms':['rms',temp.rms,0]}
                    for n,prin in enumerate(info_printed):
                        l1,l2,l3 = txt[prin]
                        if n!=0:
                            lbl = lbl+'%s'%(symb)
                        if l3!=0:
                            lbl = lbl+' '+l1+' = %.2f $\\pm$ %.2f'%(l2,l3)
                        else:
                            lbl = lbl+' '+l1+' = %.2f'%(l2)

                plt.plot(self.x,(self.x-np.mean(self.x)*int(recenter))*self.lin_slope_w+self.lin_intercept_w,color=color,ls='-.',label=lbl)
            else:
                plt.plot(self.x,(self.x-np.mean(self.x)*int(recenter))*self.lin_slope_w+self.lin_intercept_w,color=color,ls='-.')

        if info&Draw:
            plt.legend(fontsize=fontsize)

    def fit_poly(self, Draw = False, d = 2, color='r',cov=True):
        if np.sum(self.yerr)!=0:
            weights=self.yerr
        else :
            weights =np.ones(len(self.x))
        if cov:
            coeff, V = np.polyfit(self.x, self.y, d, w=1/weights,cov=cov)
            self.cov = V
            self.err = np.sqrt(np.diag(V))
        else:
            coeff= np.polyfit(self.x, self.y, d, w=1/weights,cov=cov)
        self.poly_coefficient = coeff
        self.chi2 = np.sum((self.y-np.polyval(coeff,self.x))**2)/np.sum(self.yerr**2)
        self.bic = self.chi2+(d+1)*np.log(len(self.x))
        if Draw==True:
            new_x = np.linspace(self.x.min(),self.x.max(),10000)
            plt.plot(new_x, np.polyval(coeff, new_x), linestyle='-.', color=color, linewidth=1)
            #uncertainty = np.sqrt(np.sum([(err[j]*new_x**j)**2 for j in range(len(err))],axis=0))
            #plt.fill_between(new_x,np.polyval(coeff, new_x)-uncertainty/2,np.polyval(coeff, new_x)+uncertainty/2,alpha=0.4,color=color)

    def interpolate(self, new_grid = 'auto', method = 'cubic', replace = True, interpolate_x=True, fill_value='extrapolate', scale='lin'):
        
        if scale!='lin':
            self.inv()
        
        if type(new_grid)==str:
            new_grid = np.linspace(self.x.min(),self.x.max(),10*len(self.x))
        if type(new_grid)==int:
            new_grid = np.linspace(self.x.min(),self.x.max(),new_grid*len(self.x))
        
        if np.sum(new_grid!=self.x)!=0:
            if replace:
                self.x_backup = self.x.copy()
                self.y_backup = self.y.copy()
                self.xerr_backup = self.xerr.copy()
                self.yerr_backup = self.yerr.copy()  
                self.y = interp1d(self.x, self.y, kind = method, bounds_error = False, fill_value = fill_value)(new_grid)
                if np.sum(abs(self.yerr)):
                    self.yerr = interp1d(self.x, self.yerr, kind = method, bounds_error = False, fill_value = fill_value)(new_grid)
                else:
                    self.yerr = np.zeros(len(new_grid))
                if (interpolate_x)&(bool(np.sum(abs(self.xerr)))):
                    self.xerr = interp1d(self.x, self.xerr, kind = method, bounds_error = False, fill_value = fill_value)(new_grid)  
                else:
                    self.xerr = np.zeros(len(new_grid))
                self.x = new_grid
                
                if scale!='lin':
                    self.inv()
      
            else:
                self.y_interp = interp1d(self.x, self.y, kind = method, bounds_error = False, fill_value = fill_value)(new_grid)
                if np.sum(abs(self.yerr)):
                    self.yerr_interp = interp1d(self.x, self.yerr, kind = method, bounds_error = False, fill_value = fill_value)(new_grid)
                else:
                    self.yerr_interp = np.zeros(len(new_grid))
                if (interpolate_x)&(bool(np.sum(abs(self.xerr)))):
                    self.xerr_interp = interp1d(self.x, self.xerr, kind = method, bounds_error = False, fill_value = fill_value)(new_grid)        
                else:
                    self.xerr_interp = np.zeros(len(new_grid))
                self.x_interp = new_grid
                self.interpolated = tableXY(self.x_interp,self.y_interp,self.xerr_interp,self.yerr_interp)

                if scale!='lin':
                    self.interpolated.inv()
                    self.inv()

    def masked(self,mask,replace=True):
        if replace:
            self.x = self.x[mask]
            self.y = self.y[mask]
            self.xerr = self.xerr[mask]
            self.yerr = self.yerr[mask]
        else:
            return tableXY(self.x[mask],self.y[mask],self.xerr[mask],self.yerr[mask])

    def my_bisector(self,kind='cubic',oversampling=1,weighted=True,num_outpoints='none',between_max=False,vic=10):
        """Compute the bisector of a line if the table(XY) is a table(wavelength/flux) ---"""
        self.order()
        maxi, flux = myf.local_max(self.y,vicinity=vic)        
        maxi_left = maxi[maxi<len(self.x)/2]
        flux_left = flux[maxi<len(self.x)/2]    
        maxi_right = maxi[maxi>len(self.x)/2]
        flux_right = flux[maxi>len(self.x)/2]
        
        maxi1 = 0        
        maxi2 = len(self.x)-1
        
        if between_max:
            if len(flux_left)>0:
                maxi1 = int(maxi_left[flux_left.argmax()])
            
            if len(flux_right)>0:
                maxi2 = int(maxi_right[flux_right.argmax()])
        
        selfx = self.x[maxi1:maxi2+1]
        selfy = self.y[maxi1:maxi2+1]

        if between_max:
            self.xerr = self.xerr[maxi1:maxi2+1]
            self.yerr = self.yerr[maxi1:maxi2+1]
            self.x = self.x[maxi1:maxi2+1]
            self.y = self.y[maxi1:maxi2+1]        
        normalisation = abs(selfx[selfy.argmin()-1]-selfx[selfy.argmin()+1])
        Interpol = interp1d(selfx, selfy, kind=kind, bounds_error=False, fill_value='extrapolate')
        new_x = np.linspace(selfx.min(),selfx.max(),oversampling*(len(selfx)-1)+1)
        new_y = Interpol(new_x)
        
        min_idx = new_y.argmin()
        liste1 = (new_y[0:min_idx+1])[::-1] ; liste2 = new_y[min_idx:]
        left = (new_x[0:min_idx+1])[::-1] ; right = new_x[min_idx:]
        save = myf.match_nearest(liste1,liste2)
        bisector_x = [] ; bisector_y = [] ; bisector_xerr = []
        for num in np.arange(len(save[1:-1,0]))+1:
            j = save[num,0].astype('int')
            k = save[num,1].astype('int')
            bisector_y.append(np.mean([liste1[j],liste2[k]]))
            bisector_x.append(np.mean([left[j],right[k]]))
            if weighted:
                diff_left = (liste1[j-1]-liste1[j+1])/(left[j-1]-left[j+1])
                diff_right = (liste2[k-1]-liste2[k+1])/(right[k-1]-right[k+1])
                diff = np.max([abs(diff_left),abs(diff_right)])
                bisector_xerr.append(1/diff)
                
        if weighted:
            bisector_xerr = np.array(bisector_xerr) ; bisector_xerr = bisector_xerr*normalisation/bisector_xerr[-1]/2
        else:
            bisector_xerr = np.zeros(len(bisector_y))
        bisector = np.vstack([bisector_x,bisector_y,bisector_xerr]).T
        bisector = np.insert(bisector,0,[new_x[min_idx],new_y[min_idx],np.max(bisector_xerr)],axis=0)
        Interpol = interp1d(bisector[:,1], bisector[:,0], kind=kind, bounds_error=False, fill_value='extrapolate')
        if type(num_outpoints)==str:
            num_outpoints = len(bisector[:,1]) 
        new_x = np.linspace(bisector[:,1].min(),bisector[:,1].max(),num_outpoints)
        new_y = Interpol(new_x)
        Interpol = interp1d(bisector[:,1], bisector[:,2], kind=kind, bounds_error=False, fill_value='extrapolate')
        new_yerr = Interpol(new_x)
        bisector = np.vstack([new_y,new_x,new_yerr]).T
        self.bisector = bisector.copy()
        self.bis = tableXY(bisector[:,1],bisector[:,0],bisector[:,2])

    def night_stack(self,db=0,bin_length=1,replace=False):
        
        jdb = self.x
        vrad = self.y
        vrad_std = self.yerr.copy()

        if not np.sum(vrad_std): #to avoid null vector
            vrad_std+=1    

        vrad_std[vrad_std==0] = np.nanmax(vrad_std[vrad_std!=0]*10)    
        
        weights = 1/(vrad_std)**2


        if bin_length:
            groups = ((jdb-db)//bin_length).astype('int')
            groups -= groups[0]
            group = np.unique(groups)
        else:
            group = np.arange(len(jdb))
            groups = np.arange(len(jdb))
            
        mean_jdb = []
        mean_vrad = []
        mean_svrad = []
        
        for j in group:
            g = np.where(groups==j)[0]
            mean_jdb.append(np.sum(jdb[g]*weights[g])/np.sum(weights[g]))
            mean_svrad.append(1/np.sqrt(np.sum(weights[g])))
            mean_vrad.append(np.sum(vrad[g]*weights[g])/np.sum(weights[g]))
            
        mean_jdb = np.array(mean_jdb)
        mean_vrad = np.array(mean_vrad)
        mean_svrad = np.array(mean_svrad)
        
        if replace:
            self.x, self.y, self.xerr,self.yerr = mean_jdb, mean_vrad ,0*mean_svrad, mean_svrad
        else:
            self.stacked = tableXY(mean_jdb,mean_vrad,mean_svrad)

    def order(self, order=None):
        if order is None:
            order = self.x.argsort()
        self.order_liste = order
        self.x = self.x[order]
        self.y = self.y[order]
        self.xerr = self.xerr[order]
        self.yerr = self.yerr[order]

    def plot(self, Show=False, color='k', label='', ls='', lw=2, offset=0, mask=None, capsize=3, fmt='o', markersize=6, zorder=1, species=None, alpha=1, modulo=None, modulo_norm=False, cmap=None, new=False, phase_mod=0, periodic=False, frac=1, yerr=True,xerr=True, sp=None, highlight_seasons=False,cmin=None,cmax=None):
        
        '''For the mask give either the first and last index in a list [a,b] or the mask boolean'''
        
        if modulo==100000: #default value in YARARA
            modulo=None
        
        if (modulo is not None)&(cmap is None):
            try:
                cmap = {'k':'viridis','b':'Blues','r':'Reds','g':'Greens'}[color]
            except:
                cmap = 'viridis'
        if (len(self.x)>20000)&(ls=='')&(modulo is None):
            ls='-'
        
        if species is None:
            species = np.ones(len(self.x))

        if highlight_seasons:
            self.split_seasons(min_gap=highlight_seasons,Plot=False)
            species = self.seasons_species

        if len(np.unique(species))==1:
            colors_species = [color]
        else:
            colors_species = ['k']+['C%.0f'%(i) for i in range(1,1+len(np.unique(species)))]

        for num, selection in enumerate(np.unique(species)):
            
            color = colors_species[num]
            
            if num!=0:
                label=None
            
            if mask is None:
                mask2 = np.ones(len(self.x)).astype('bool')
            elif type(mask[0])==int:
                mask2 = np.zeros(len(self.x)).astype('bool')
                mask2[mask[0]:mask[1]]=True
            else:
                mask2 = mask

            loc = np.where(species[mask2]==selection)[0]
            
            sel = np.arange(len(loc))
            if frac!=1:
                sel = np.random.choice(np.arange(len(loc)),size=int(frac*len(loc)),replace=False)
                
            if new:
                plt.figure()
                
            if sp is not None:
                plt.subplot(sp)
            
            if ls!='':
                if color is not None:
                    if len(color)==len(self.x):
                        points   = np.array([self.x, self.y]).T.reshape(-1, 1, 2)
                        segments = np.concatenate([points[:-1], points[1:]], axis=1)
                        if cmin is None:
                            cmin = np.nanpercentile(color,5)
                        if cmax is None:
                            cmax = np.nanpercentile(color,95)
                        norm = plt.Normalize(cmin, cmax)    
                        lc = LineCollection(segments, cmap=cmap, norm=norm,lw=lw)
                        lc.set_array(color)
                        line = plt.gca().add_collection(lc)
                        #axc = fig.colorbar(line, ax=plt.gca())
                    else:
                        plt.plot(self.x[mask2][loc][sel],self.y[mask2][loc][sel]+offset,ls=ls,lw=lw,zorder=zorder,label=label,color=color,alpha=alpha)
                else:
                    plt.plot(self.x[mask2][loc][sel],self.y[mask2][loc][sel]+offset,ls=ls,lw=lw,zorder=zorder,label=label,color=color,alpha=alpha)
            else:
                if modulo is not None:
                    norm = ((1-modulo_norm)+modulo_norm*modulo)
                    new_x = ((self.x[mask2][loc]-phase_mod)%modulo)/norm
                    plt.errorbar(new_x[sel], self.y[mask2][loc][sel]+offset, xerr=self.xerr[mask2][loc][sel], yerr=self.yerr[mask2][loc][sel]*int(yerr), fmt=fmt, color=color, alpha=alpha, capsize=capsize, label=label, markersize=markersize,zorder=zorder)
                    plt.scatter(new_x[sel], self.y[mask2][loc][sel]+offset, marker='o', c=self.x[mask2][loc][sel], cmap=cmap, s=markersize,zorder=zorder*100,vmin=np.nanpercentile(self.x[mask2][loc][sel],16),vmax=np.nanpercentile(self.x[mask2][loc][sel],84))
                    if periodic:
                        for i in range(int(np.float(periodic))):
                            plt.errorbar(new_x[sel] - (i+1)*modulo/norm, self.y[mask2][loc][sel]+offset, xerr=self.xerr[mask2][loc][sel], yerr=self.yerr[mask2][loc][sel], fmt=fmt, color=color, alpha=0.3, capsize=capsize, markersize=markersize, zorder=zorder)
                            plt.errorbar(new_x[sel] + (i+1)*modulo/norm, self.y[mask2][loc][sel]+offset, xerr=self.xerr[mask2][loc][sel], yerr=self.yerr[mask2][loc][sel], fmt=fmt, color=color, alpha=0.3, capsize=capsize, markersize=markersize, zorder=zorder)
                else:
                    plt.errorbar(self.x[mask2][loc][sel], self.y[mask2][loc][sel]+offset, xerr=self.xerr[mask2][loc][sel]*int(xerr), yerr=self.yerr[mask2][loc][sel]*int(yerr), fmt=fmt, color=color, alpha=alpha, capsize=capsize, label=label, markersize=markersize,zorder=zorder)
        if Show==True:
            plt.legend()
            plt.show()

    def recenter(self, who='both',weight=False):
        if (who=='X')|(who=='both')|(who=='x'):
            self.xmean = np.nanmean(self.x)
            self.x = self.x - np.nanmean(self.x)
        if (who=='Y')|(who=='both')|(who=='y'):
            self.ymean = np.nanmean(self.y)
            self.y = self.y - np.nanmean(self.y)

    def rm_outliers(self, who='Y', m=2, kind='inter', bin_length=0, replace=True):
        vec = self.copy()
        if bin_length:
            self.night_stack(bin_length=bin_length,replace=False)
            vec_binned = self.stacked.copy()
        else:
            vec_binned = self.copy()
            
        if who=='Xerr':
            mask = rm_out(self.xerr, m=m, kind=kind)[0]
            vec_binned = vec_binned.xerr
            vec = vec.xerr
        if who=='Yerr':
            mask = rm_out(self.yerr, m=m, kind=kind)[0]
            vec_binned = vec_binned.yerr
            vec = vec.yerr
        if who=='Y':
            mask = rm_out(self.y, m=m, kind=kind)[0]
            vec_binned = vec_binned.y
            vec = vec.y
        if who=='X':
           mask = rm_out(self.x, m=m, kind=kind)[0]
           vec_binned = vec_binned.x
           vec = vec.x
         
        if bin_length:
            outputs = rm_out(vec_binned,m=m,kind=kind,return_borders=True)
            mask = (vec>=outputs[-1])&(vec<=outputs[-2])
           
        if who =='both':
            mask = rm_out(self.x, m=m, kind=kind)[0] & rm_out(self.y, m=m, kind=kind)[0] 
        self.mask = mask
        if replace:
            self.x = self.x[mask]
            self.y = self.y[mask]
            self.yerr = self.yerr[mask]
            self.xerr = self.xerr[mask]

    def rms_w(self):
        if len(self.x)>1:
            self.rms = DescrStatsW(self.y,weights=1./self.yerr**2).std
            self.weighted_average = DescrStatsW(self.y,weights=1./self.yerr**2).mean
            self.stats['rms'] = self.rms
        else:
            self.rms=0
            self.weighted_average = self.y[0]


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 13 20:26:21 2018

@author: allartromain
"""

import numpy as np
import pylab as pl
import os
from astropy.io import fits 
import matplotlib as mpl
import time
from lmfit import Model


def find_nearest(array, values):
    array = np.asarray(array)

    # the last dim must be 1 to broadcast in (array - values) below.
    values = np.expand_dims(values, axis=-1) 

    indices = np.abs(array - values).argmin(axis=-1)

    return indices

def read_hitran2012_parfile(filename):
    '''
    Given a HITRAN2012-format text file, read in the parameters of the molecular absorption features.
    Parameters
    ----------
    filename : str
        The filename to read in.
    Return
    ------
    data : dict
        The dictionary of HITRAN data for the molecule.
    '''

    if not os.path.exists:
        raise ImportError('The input filename"' + filename + '" does not exist.')

    if filename.endswith('.zip'):
        import zipfile
        zip = zipfile.ZipFile(filename, 'r')
        (object_name, ext) = os.path.splitext(os.path.basename(filename))
        print((object_name, ext))
        filehandle = zip.read(object_name).splitlines()
    else:
        filehandle = open(filename, 'r')

    data = {'M': [],  ## molecule identification number
            'I': [],  ## isotope number
            'linecenter': [],  ## line center wavenumber (in cm^{-1})
            'S': [],  ## line strength, in cm^{-1} / (molecule m^{-2})
            'Acoeff': [],  ## Einstein A coefficient (in s^{-1})
            'gamma-air': [],  ## line HWHM for air-broadening
            'gamma-self': [],  ## line HWHM for self-emission-broadening
            'Epp': [],  ## energy of lower transition level (in cm^{-1})
            'N': [],  ## temperature-dependent exponent for "gamma-air"
            'delta': [],  ## air-pressure shift, in cm^{-1} / atm
            'Vp': [],  ## upper-state "global" quanta index
            'Vpp': [],  ## lower-state "global" quanta index
            'Qp': [],  ## upper-state "local" quanta index
            'Qpp': [],  ## lower-state "local" quanta index
            'Ierr': [],  ## uncertainty indices
            'Iref': [],  ## reference indices
            'flag': [],  ## flag
            'gp': [],  ## statistical weight of the upper state
            'gpp': []}  ## statistical weight of the lower state

    print(('Reading "' + filename + '" ...'))

    for line in filehandle:
        if (len(line) < 160):
            raise ImportError(
                'The imported file ("' + filename + '") does not appear to be a HITRAN2012-format data file.')

        data['M'].append(np.uint(line[0:2]))
        data['I'].append(np.uint(line[2]))
        data['linecenter'].append(np.float64(line[3:15]))
        data['S'].append(np.float64(line[15:25]))
        data['Acoeff'].append(np.float64(line[25:35]))
        data['gamma-air'].append(np.float64(line[35:40]))
        data['gamma-self'].append(np.float64(line[40:45]))
        data['Epp'].append(np.float64(line[45:55]))
        data['N'].append(np.float64(line[55:59]))
        data['delta'].append(np.float64(line[59:67]))
        data['Vp'].append(line[67:82])
        data['Vpp'].append(line[82:97])
        data['Qp'].append(line[97:112])
        data['Qpp'].append(line[112:127])
        data['Ierr'].append(line[127:133])
        data['Iref'].append(line[133:145])
        data['flag'].append(line[145])
        data['gp'].append(line[146:153])
        data['gpp'].append(line[153:160])

    if filename.endswith('.zip'):
        zip.close()
    else:
        filehandle.close()

    for key in data:
        data[key] = np.array(data[key])

    return (data)

"""
def static_hitran_file(hitran_file, QT_file, species, S_cut=10 ** -26,instrument='ESPRESSO'):
    hitran_database = read_hitran2012_parfile(hitran_file)
    qt_file = np.genfromtxt(QT_file)
    
    wave_number = hitran_database['linecenter'][hitran_database['S'] > S_cut]
    Intensity   = hitran_database['S'][hitran_database['S'] > S_cut]
    gamma_air   = hitran_database['gamma-air'][hitran_database['S'] > S_cut]
    gamma_self  = hitran_database['gamma-self'][hitran_database['S'] > S_cut]
    N           = hitran_database['N'][hitran_database['S'] > S_cut]
    delta       = hitran_database['delta'][hitran_database['S'] > S_cut]
    Epp         = hitran_database['Epp'][hitran_database['S'] > S_cut]
    
    removed_lines = [13570.9,13704.525,13708.60,13708.625,13766.465,13941.252,13947.17,13928.5,15437.85,15437.89,15437.9,
                     13701.8,13910.16]
    
    
    col1 = fits.Column(name='wave_number', format='1D', array=np.delete(wave_number,find_nearest(wave_number, removed_lines)))
    col2 = fits.Column(name='Intensity',   format='1D', array=np.delete(Intensity,find_nearest(wave_number, removed_lines)))
    col3 = fits.Column(name='gamma_air',   format='1D', array=np.delete(gamma_air,find_nearest(wave_number, removed_lines)))
    col3b = fits.Column(name='gamma_self', format='1D', array=np.delete(gamma_self,find_nearest(wave_number, removed_lines)))
    col4 = fits.Column(name='N',           format='1D', array=np.delete(N,find_nearest(wave_number, removed_lines)))
    col5 = fits.Column(name='delta',       format='1D', array=np.delete(delta,find_nearest(wave_number, removed_lines)))
    col6 = fits.Column(name='Epp',         format='1D', array=np.delete(Epp,find_nearest(wave_number, removed_lines)))

    col7 = fits.Column(name='Temperature', format='1D', array=qt_file[:, 0])
    col8 = fits.Column(name='Qt', format='1D', array=qt_file[:, 1])

    cols0 = fits.ColDefs([col1, col2, col3, col3b, col4, col5, col6])
    tbhdu0 = fits.BinTableHDU.from_columns(cols0)

    cols1 = fits.ColDefs([col7, col8])
    tbhdu1 = fits.BinTableHDU.from_columns(cols1)
    prihdr = fits.Header()
    prihdr['Intensity Cut'] = str(S_cut)
    prihdu = fits.PrimaryHDU(header=prihdr)
    thdulist = fits.HDUList([prihdu, tbhdu0, tbhdu1])
    thdulist.writeto('Static_model/'+instrument+'/Static_hitran_qt_'+species+'.fits',
                     overwrite=True)
    return
"""
#(tavella)
def static_hitran_file(hitran_file, QT_file, species, S_cut=10 ** -26,instrument='HARPS_N'):
    hitran_database = read_hitran2012_parfile(hitran_file)
    qt_file = np.genfromtxt(QT_file)
    
    wave_number = hitran_database['linecenter'][hitran_database['S'] > S_cut]
    Intensity   = hitran_database['S'][hitran_database['S'] > S_cut]
    gamma_air   = hitran_database['gamma-air'][hitran_database['S'] > S_cut]
    gamma_self  = hitran_database['gamma-self'][hitran_database['S'] > S_cut]
    N           = hitran_database['N'][hitran_database['S'] > S_cut]
    delta       = hitran_database['delta'][hitran_database['S'] > S_cut]
    Epp         = hitran_database['Epp'][hitran_database['S'] > S_cut]
    
    removed_lines = [13570.9,13704.525,13708.60,13708.625,13766.465,13941.252,13947.17,13928.5,15437.85,15437.89,15437.9,
                     13701.8,13910.16]
    
    
    col1 = fits.Column(name='wave_number', format='1D', array=np.delete(wave_number,find_nearest(wave_number, removed_lines)))
    col2 = fits.Column(name='Intensity',   format='1D', array=np.delete(Intensity,find_nearest(wave_number, removed_lines)))
    col3 = fits.Column(name='gamma_air',   format='1D', array=np.delete(gamma_air,find_nearest(wave_number, removed_lines)))
    col3b = fits.Column(name='gamma_self', format='1D', array=np.delete(gamma_self,find_nearest(wave_number, removed_lines)))
    col4 = fits.Column(name='N',           format='1D', array=np.delete(N,find_nearest(wave_number, removed_lines)))
    col5 = fits.Column(name='delta',       format='1D', array=np.delete(delta,find_nearest(wave_number, removed_lines)))
    col6 = fits.Column(name='Epp',         format='1D', array=np.delete(Epp,find_nearest(wave_number, removed_lines)))

    col7 = fits.Column(name='Temperature', format='1D', array=qt_file[:, 0])
    col8 = fits.Column(name='Qt', format='1D', array=qt_file[:, 1])

    cols0 = fits.ColDefs([col1, col2, col3, col3b, col4, col5, col6])
    tbhdu0 = fits.BinTableHDU.from_columns(cols0)

    cols1 = fits.ColDefs([col7, col8])
    tbhdu1 = fits.BinTableHDU.from_columns(cols1)
    prihdr = fits.Header()
    prihdr['Intensity Cut'] = str(S_cut)
    prihdu = fits.PrimaryHDU(header=prihdr)
    thdulist = fits.HDUList([prihdu, tbhdu0, tbhdu1])
    thdulist.writeto('Static_model/'+instrument+'/Static_hitran_qt_'+species+'.fits',
                     overwrite=True)
    return


"""

def lines_to_fit(input_file,species,wavenumber_born=np.array([10**8/7000,10**8/8000]),number_of_lines=10,region_excluded=np.array([10**8/12680,10**8/12710]),lines_excluded=np.array([1.39149595e+04,1.38726811e+04,1.38184074e+04]),instrument='ESPRESSO'):
    "wave_born, region_excluded and lines_excluded are in wavenumber in cm-1"
    hitran_database = fits.open(input_file)[1].data
    
    wave_number_hitran_rest_temp = hitran_database['wave_number'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    intensity_hitran_temp        = hitran_database['Intensity'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    gamma_air_temp               = hitran_database['gamma_air'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    n_air_temp                   = hitran_database['N'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    delta_air_temp               = hitran_database['delta'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    Epp_temp                     = hitran_database['Epp'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
        
   	##############################################
   	####    Selection of the strongest lines  ####
   	##############################################

    wave_number_hitran_rest_Nstrong          = []
    wave_number_hitran_rest_temp_order       = wave_number_hitran_rest_temp[np.argsort(intensity_hitran_temp)]
    wave_number_hitran_rest_excluded         = []


    for sw in range(len(wave_number_hitran_rest_temp_order)):
        idx = -1-sw
        #if (np.round(wave_number_hitran_rest_temp_order[idx],4) not in np.round(lines_excluded,4)) & ((wave_number_hitran_rest_temp_order[idx] > region_excluded[0]) | (wave_number_hitran_rest_temp_order[idx] < region_excluded[1])):
        if (np.sum((wave_number_hitran_rest_temp_order[idx] < 10**8/(10**8/lines_excluded + 2*40./300000*np.mean(10**8/lines_excluded))) | (wave_number_hitran_rest_temp_order[idx] > 10**8/(10**8/lines_excluded - 2*40./300000*np.mean(10**8/lines_excluded))))==len(lines_excluded)) & ((wave_number_hitran_rest_temp_order[idx] > region_excluded[0]) | (wave_number_hitran_rest_temp_order[idx] < region_excluded[1])):
            if (len(wave_number_hitran_rest_Nstrong) == 0):
                wave_number_hitran_rest_Nstrong = np.concatenate((wave_number_hitran_rest_Nstrong,[wave_number_hitran_rest_temp_order[idx]]))
            elif (len(wave_number_hitran_rest_Nstrong) != 0)  :
                rejection_criteria=0
                for swl in range(len(wave_number_hitran_rest_Nstrong)):
                    if (wave_number_hitran_rest_temp_order[idx] > 10**8/(10**8/wave_number_hitran_rest_Nstrong[swl] + 2*40./300000*np.mean(10**8/wavenumber_born))) and (wave_number_hitran_rest_temp_order[idx] < 10**8/(10**8/wave_number_hitran_rest_Nstrong[swl] - 2*40./300000*np.mean(10**8/wavenumber_born)))  :  #avoid two strong telluric line blended
                        rejection_criteria+=1
                if (rejection_criteria ==0) :
                    wave_number_hitran_rest_Nstrong = np.concatenate((wave_number_hitran_rest_Nstrong,[wave_number_hitran_rest_temp_order[idx]]))
                else:
                    wave_number_hitran_rest_excluded = np.concatenate((wave_number_hitran_rest_excluded,[wave_number_hitran_rest_temp_order[idx]]))
            if len(wave_number_hitran_rest_Nstrong) >=number_of_lines:
                break
        else:
            wave_number_hitran_rest_excluded = np.concatenate((wave_number_hitran_rest_excluded,[wave_number_hitran_rest_temp_order[idx]]))

    wave_number_hitran_rest_Nstrong  = np.array(wave_number_hitran_rest_Nstrong)
    wave_number_hitran_rest_excluded = np.array(wave_number_hitran_rest_excluded)


    wave_number_hitran_rest_Nstrong_below = 1./((1./(wave_number_hitran_rest_Nstrong*10**-8) - 40./300000*np.mean(10**8/wavenumber_born)) * 10**-8 )
    wave_number_hitran_rest_Nstrong_upper = 1./((1./(wave_number_hitran_rest_Nstrong*10**-8) + 40./300000*np.mean(10**8/wavenumber_born)) * 10**-8 )

   	##################################################################
   	####    Selection of all the lines around the strongest ones  ####
   	##################################################################

    wave_number_hitran_rest = []
    intensity_hitran        = []
    gamma_air               = []
    n_air                   = []
    delta_air               = []
    Epp                     = []

    for wn in range(len(wave_number_hitran_rest_Nstrong)): # select all the lines around the N strongest
        wave_number_hitran_rest = np.concatenate(( wave_number_hitran_rest, wave_number_hitran_rest_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        intensity_hitran        = np.concatenate(( intensity_hitran,        intensity_hitran_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        gamma_air               = np.concatenate(( gamma_air,               gamma_air_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        n_air                   = np.concatenate(( n_air,                   n_air_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        delta_air               = np.concatenate(( delta_air,               delta_air_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        Epp                     = np.concatenate(( Epp,                     Epp_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))



    col1 = fits.Column(name='wave_number', format='1D', array=wave_number_hitran_rest)
    col2 = fits.Column(name='Intensity',   format='1D', array=intensity_hitran)
    col3 = fits.Column(name='gamma_air',   format='1D', array=gamma_air)
    col4 = fits.Column(name='N',           format='1D', array=n_air)
    col5 = fits.Column(name='delta',       format='1D', array=delta_air)
    col6 = fits.Column(name='Epp',         format='1D', array=Epp)

    col7 = fits.Column(name='CCF_lines_position_wavenumber', format='1D', array=wave_number_hitran_rest_Nstrong)
    col8 = fits.Column(name='CCF_lines_position_wavelength', format='1D', array=10**8/wave_number_hitran_rest_Nstrong)

    col9 = fits.Column(name='Excluded_lines_position_wavenumber', format='1D', array=wave_number_hitran_rest_excluded)
    col10 = fits.Column(name='Excluded_lines_position_wavelength', format='1D', array=10**8/wave_number_hitran_rest_excluded)

    cols0 = fits.ColDefs([col1, col2, col3, col4, col5, col6])
    tbhdu0 = fits.BinTableHDU.from_columns(cols0)

    cols1 = fits.ColDefs([col7, col8])
    tbhdu1 = fits.BinTableHDU.from_columns(cols1)

    cols2 = fits.ColDefs([col9, col10])
    tbhdu2 = fits.BinTableHDU.from_columns(cols2)    
    
    prihdr = fits.Header()
    prihdr['Number of lines'] = str(number_of_lines)
    prihdu = fits.PrimaryHDU(header=prihdr)
    thdulist = fits.HDUList([prihdu, tbhdu0, tbhdu1, tbhdu2])
    thdulist.writeto('Static_model/'+instrument+'/Static_hitran_strongest_lines_'+species+'.fits', overwrite=True)

    return

"""
#(tavella)
def lines_to_fit(input_file,species,wavenumber_born=np.array([10**8/7000,10**8/8000]),number_of_lines=10,region_excluded=np.array([10**8/12680,10**8/12710]),lines_excluded=np.array([1.39149595e+04,1.38726811e+04,1.38184074e+04]),instrument='HARPS_N'):
    "wave_born, region_excluded and lines_excluded are in wavenumber in cm-1"
    hitran_database = fits.open(input_file)[1].data
    
    wave_number_hitran_rest_temp = hitran_database['wave_number'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    intensity_hitran_temp        = hitran_database['Intensity'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    gamma_air_temp               = hitran_database['gamma_air'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    n_air_temp                   = hitran_database['N'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    delta_air_temp               = hitran_database['delta'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
    Epp_temp                     = hitran_database['Epp'][(wavenumber_born[-1] < hitran_database['wave_number']) &  (hitran_database['wave_number'] < wavenumber_born[0])]
        
   	##############################################
   	####    Selection of the strongest lines  ####
   	##############################################

    wave_number_hitran_rest_Nstrong          = []
    delta_air_temp_Nstrong                   = []
    wave_number_hitran_rest_temp_order       = wave_number_hitran_rest_temp[np.argsort(intensity_hitran_temp)]
    delta_air_temp_order                     = delta_air_temp[np.argsort(intensity_hitran_temp)]
    wave_number_hitran_rest_excluded         = []


    for sw in range(len(wave_number_hitran_rest_temp_order)):
        idx = -1-sw
        #if (np.round(wave_number_hitran_rest_temp_order[idx],4) not in np.round(lines_excluded,4)) & ((wave_number_hitran_rest_temp_order[idx] > region_excluded[0]) | (wave_number_hitran_rest_temp_order[idx] < region_excluded[1])):
        if (np.sum((wave_number_hitran_rest_temp_order[idx] < 10**8/(10**8/lines_excluded + 2*40./300000*np.mean(10**8/lines_excluded))) | (wave_number_hitran_rest_temp_order[idx] > 10**8/(10**8/lines_excluded - 2*40./300000*np.mean(10**8/lines_excluded))))==len(lines_excluded)) & ((wave_number_hitran_rest_temp_order[idx] > region_excluded[0]) | (wave_number_hitran_rest_temp_order[idx] < region_excluded[1])):
            if (len(wave_number_hitran_rest_Nstrong) == 0):
                wave_number_hitran_rest_Nstrong = np.concatenate((wave_number_hitran_rest_Nstrong,[wave_number_hitran_rest_temp_order[idx]]))
                delta_air_temp_Nstrong = np.concatenate((delta_air_temp_Nstrong,[delta_air_temp_order[idx]]))
            elif (len(wave_number_hitran_rest_Nstrong) != 0)  :
                rejection_criteria=0
                for swl in range(len(wave_number_hitran_rest_Nstrong)):
                    if (wave_number_hitran_rest_temp_order[idx] > 10**8/(10**8/wave_number_hitran_rest_Nstrong[swl] + 2*40./300000*np.mean(10**8/wavenumber_born))) and (wave_number_hitran_rest_temp_order[idx] < 10**8/(10**8/wave_number_hitran_rest_Nstrong[swl] - 2*40./300000*np.mean(10**8/wavenumber_born)))  :  #avoid two strong telluric line blended
                        rejection_criteria+=1
                if (rejection_criteria ==0) :
                    wave_number_hitran_rest_Nstrong = np.concatenate((wave_number_hitran_rest_Nstrong,[wave_number_hitran_rest_temp_order[idx]]))
                    delta_air_temp_Nstrong = np.concatenate((delta_air_temp_Nstrong,[delta_air_temp_order[idx]]))
                else:
                    wave_number_hitran_rest_excluded = np.concatenate((wave_number_hitran_rest_excluded,[wave_number_hitran_rest_temp_order[idx]]))
            if len(wave_number_hitran_rest_Nstrong) >=number_of_lines:
                break
        else:
            wave_number_hitran_rest_excluded = np.concatenate((wave_number_hitran_rest_excluded,[wave_number_hitran_rest_temp_order[idx]]))

    wave_number_hitran_rest_Nstrong  = np.array(wave_number_hitran_rest_Nstrong)
    delta_air_temp_Nstrong  = np.array(delta_air_temp_Nstrong)
    wave_number_hitran_rest_excluded = np.array(wave_number_hitran_rest_excluded)


    wave_number_hitran_rest_Nstrong_below = 1./((1./(wave_number_hitran_rest_Nstrong*10**-8) - 40./300000*np.mean(10**8/wavenumber_born)) * 10**-8 )
    wave_number_hitran_rest_Nstrong_upper = 1./((1./(wave_number_hitran_rest_Nstrong*10**-8) + 40./300000*np.mean(10**8/wavenumber_born)) * 10**-8 )

   	##################################################################
   	####    Selection of all the lines around the strongest ones  ####
   	##################################################################

    wave_number_hitran_rest = []
    intensity_hitran        = []
    gamma_air               = []
    n_air                   = []
    delta_air               = []
    Epp                     = []

    for wn in range(len(wave_number_hitran_rest_Nstrong)): # select all the lines around the N strongest
        wave_number_hitran_rest = np.concatenate(( wave_number_hitran_rest, wave_number_hitran_rest_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        intensity_hitran        = np.concatenate(( intensity_hitran,        intensity_hitran_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        gamma_air               = np.concatenate(( gamma_air,               gamma_air_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        n_air                   = np.concatenate(( n_air,                   n_air_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        delta_air               = np.concatenate(( delta_air,               delta_air_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))
        Epp                     = np.concatenate(( Epp,                     Epp_temp[(wave_number_hitran_rest_Nstrong_below[wn] > wave_number_hitran_rest_temp) &  (wave_number_hitran_rest_temp > wave_number_hitran_rest_Nstrong_upper[wn])] ))



    col1 = fits.Column(name='wave_number', format='1D', array=wave_number_hitran_rest)
    col2 = fits.Column(name='Intensity',   format='1D', array=intensity_hitran)
    col3 = fits.Column(name='gamma_air',   format='1D', array=gamma_air)
    col4 = fits.Column(name='N',           format='1D', array=n_air)
    col5 = fits.Column(name='delta',       format='1D', array=delta_air)
    col6 = fits.Column(name='Epp',         format='1D', array=Epp)

    col7 = fits.Column(name='CCF_lines_position_wavenumber', format='1D', array=wave_number_hitran_rest_Nstrong)
    col8 = fits.Column(name='CCF_lines_position_wavelength', format='1D', array=10**8/wave_number_hitran_rest_Nstrong)
    col8bis = fits.Column(name='CCF_lines_delta', format='1D', array=delta_air_temp_Nstrong)

    col9 = fits.Column(name='Excluded_lines_position_wavenumber', format='1D', array=wave_number_hitran_rest_excluded)
    col10 = fits.Column(name='Excluded_lines_position_wavelength', format='1D', array=10**8/wave_number_hitran_rest_excluded)

    cols0 = fits.ColDefs([col1, col2, col3, col4, col5, col6])
    tbhdu0 = fits.BinTableHDU.from_columns(cols0)

    cols1 = fits.ColDefs([col7, col8 , col8bis])
    tbhdu1 = fits.BinTableHDU.from_columns(cols1)

    cols2 = fits.ColDefs([col9, col10])
    tbhdu2 = fits.BinTableHDU.from_columns(cols2)    
    
    prihdr = fits.Header()
    prihdr['Number of lines'] = str(number_of_lines)
    prihdu = fits.PrimaryHDU(header=prihdr)
    thdulist = fits.HDUList([prihdu, tbhdu0, tbhdu1, tbhdu2])
    thdulist.writeto('Static_model/'+instrument+'/Static_hitran_strongest_lines_'+species+'.fits', overwrite=True)

    return



def polynomial_regression_2d(xy, a, b, c, d, e, f, g):
    x = xy[0]
    y = xy[1]
    return a * x + b * y + c * x * y + d * x ** 2. + e * y ** 2. + f * x ** 2. * y ** 2. + g

def shape_s2d_ESPRESSO(time_science,ins_mode,bin_x):
    if (ins_mode == 'SINGLEUHR') or (ins_mode == 'SINGLEHR'):
        if time_science < 58421.5:
            index_pixel_blue = np.arange(1361., 7936. + 1, 1)
            index_pixel_red = np.arange(1., 9211. + 1, 1)
        elif (time_science > 58421.5) and (time_science < 58653.1):
            index_pixel_blue = np.arange(1361., 7901. + 1, 1)
            index_pixel_red = np.arange(1., 9111. + 1, 1)
        elif time_science > 58653.1:
            index_pixel_blue = np.arange(1361., 7901. + 1, 1)
            index_pixel_red = np.arange(1., 9111. + 1, 1)

    elif (ins_mode == 'MULTIMR') and (bin_x == 4):
        if time_science < 58421.5:
            index_pixel_blue = np.arange(591., 4040. + 1, 1)
            index_pixel_red = np.arange(1., 4590. + 1, 1)
        elif (time_science > 58421.5) and (time_science < 58653.1):
            index_pixel_blue = np.arange(691., 3930. + 1, 1)
            index_pixel_red = np.arange(1., 4545. + 1, 1)
        elif time_science > 58653.1:
            index_pixel_blue = np.arange(691., 3930. + 1, 1)
            index_pixel_red = np.arange(1., 4545. + 1, 1)

    elif (ins_mode == 'MULTIMR') and (bin_x == 8):
        if time_science < 58421.5:
            index_pixel_blue = np.arange(296., 2020. + 1, 1)
            index_pixel_red = np.arange(1., 2295. + 1, 1)
        elif (time_science > 58421.5) and (time_science < 58653.1):
            index_pixel_blue = np.arange(346., 1965. + 1, 1)
            index_pixel_red = np.arange(1., 2275. + 1, 1)
        elif time_science > 58653.1:
            index_pixel_blue = np.arange(346., 1965. + 1, 1)
            index_pixel_red = np.arange(1., 2275. + 1, 1)

    return index_pixel_blue, index_pixel_red


def plot_resolution_map_UHR_HR(pixels_blue_slice1,orders_blue_slice1,resolution_blue_slice1,pixels_blue_hr,matrix_resolution_blue_hr_theo_slice1,pixels_red_slice1,
                               orders_red_slice1,resolution_red_slice1,pixels_red_hr,matrix_resolution_red_hr_theo_slice1,res_min,res_max,
                               pixels_blue_slice2,orders_blue_slice2,resolution_blue_slice2,matrix_resolution_blue_hr_theo_slice2,pixels_red_slice2,
                               orders_red_slice2,resolution_red_slice2,matrix_resolution_red_hr_theo_slice2,ins_mode,binx,mjd,Doit=False):
    if Doit==True:
        fig = pl.figure(figsize=(20, 20))
        pl.scatter(pixels_blue_slice1, orders_blue_slice1, c=resolution_blue_slice1, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_blue_hr, np.arange(1, 90.01, 1), matrix_resolution_blue_hr_theo_slice1, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.scatter(pixels_red_slice1, orders_red_slice1, c=resolution_red_slice1, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_red_hr, np.arange(91, 170.01, 1), matrix_resolution_red_hr_theo_slice1, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.scatter(pixels_blue_slice2, orders_blue_slice2, c=resolution_blue_slice2, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_blue_hr, np.arange(1, 90.01, 1), matrix_resolution_blue_hr_theo_slice2, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.scatter(pixels_red_slice2, orders_red_slice2, c=resolution_red_slice2, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_red_hr, np.arange(91, 170.01, 1), matrix_resolution_red_hr_theo_slice2, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.xlabel('Pixels', size=30)
        pl.ylabel('Orders', size=30)
        pl.xticks(size=20)
        pl.yticks(size=20)
        pl.ylim(0, 170)
        pl.xlim(0, 9211)
        cbar_ax = fig.add_axes([0.93, 0.11, 0.014, 0.873])
        axcb = pl.colorbar(cax=cbar_ax, format='%0.0f')
        axcb.ax.tick_params(labelsize=20)
        axcb.set_label('Resolution', size=30)
        pl.tight_layout(w_pad=0.1, h_pad=2, rect=(0.0005, 0.0001, 0.935, 0.999))
        pl.savefig('Output/Resolution_map/ESPRESSO/Map_resolution_' + ins_mode + '_' + str(binx) + '_' + str(int(mjd)) + '.png', dpi=100, format='png', bbox_inches='tight')
    return

def plot_resolution_map_slice_UHR_HR(pixels_blue_slice1,orders_blue_slice1,resolution_blue_slice1,pixels_blue_hr,matrix_resolution_blue_hr_theo_slice1,pixels_red_slice1,
                                     orders_red_slice1,resolution_red_slice1,pixels_red_hr,matrix_resolution_red_hr_theo_slice1,res_min,res_max,
                                     pixels_blue_slice2,orders_blue_slice2,resolution_blue_slice2,matrix_resolution_blue_hr_theo_slice2,pixels_red_slice2,
                                     orders_red_slice2,resolution_red_slice2,matrix_resolution_red_hr_theo_slice2,ins_mode,binx,mjd,Doit=False):
    if Doit == True:
        fig = pl.figure(figsize=(20, 20))
        pl.subplot(1, 2, 1)
        pl.scatter(pixels_blue_slice1, orders_blue_slice1, c=resolution_blue_slice1, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_blue_hr, np.arange(1, 90.01, 1), matrix_resolution_blue_hr_theo_slice1, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.scatter(pixels_red_slice1, orders_red_slice1, c=resolution_red_slice1, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_red_hr, np.arange(91, 170.01, 1), matrix_resolution_red_hr_theo_slice1, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.xlabel('Pixels', size=30)
        pl.ylabel('Orders', size=30)
        pl.xticks(size=20)
        pl.yticks(size=20)
        pl.ylim(0, 170)
        pl.xlim(0, 9211)
        cbar_ax = fig.add_axes([0.93, 0.092, 0.014, 0.89])
        axcb = pl.colorbar(cax=cbar_ax, format='%0.2f')
        axcb.set_label('Resolution', size=30)
        pl.tight_layout(w_pad=0.1, h_pad=2, rect=(0.0005, 0.0001, 0.935, 0.999))
        pl.subplot(1, 2, 2)
        pl.scatter(pixels_blue_slice2, orders_blue_slice2, c=resolution_blue_slice2, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_blue_hr, np.arange(1, 90.01, 1), matrix_resolution_blue_hr_theo_slice2, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.scatter(pixels_red_slice2, orders_red_slice2, c=resolution_red_slice2, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_red_hr, np.arange(91, 170.01, 1), matrix_resolution_red_hr_theo_slice2, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.xlabel('Pixels', size=30)
        pl.ylabel('Orders', size=30)
        pl.xticks(size=20)
        pl.yticks(size=20)
        pl.ylim(0, 170)
        pl.xlim(0, 9211)
        cbar_ax = fig.add_axes([0.93, 0.092, 0.014, 0.89])
        axcb = pl.colorbar(cax=cbar_ax, format='%0.0f')
        axcb.ax.tick_params(labelsize=20)
        axcb.set_label('Resolution', size=30)
        pl.tight_layout(w_pad=0.1, h_pad=2, rect=(0.0005, 0.0001, 0.935, 0.999))
        pl.savefig('Output/Resolution_map/ESPRESSO/Map_resolution_slices_' + ins_mode + '_' + str(binx) + '_' + str(int(mjd)) + '.png', dpi=100, format='png', bbox_inches='tight')
    return

def plot_resolution_map_MR(pixels_blue_slice1,orders_blue_slice1,resolution_blue_slice1,pixels_blue_hr,matrix_resolution_blue_hr_theo_slice1,pixels_red_slice1,
                           orders_red_slice1,resolution_red_slice1,pixels_red_hr,matrix_resolution_red_hr_theo_slice1,res_min,res_max,
                           ins_mode,binx,mjd,Doit=False):
    if Doit == True:
        fig = pl.figure(figsize=(20, 20))
        pl.scatter(pixels_blue_slice1, orders_blue_slice1, c=resolution_blue_slice1, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_blue_hr, np.arange(1, 45.01, 1), matrix_resolution_blue_hr_theo_slice1, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.scatter(pixels_red_slice1, orders_red_slice1, c=resolution_red_slice1, cmap=mpl.cm.rainbow, vmin=res_min,vmax=res_max, edgecolors='white', zorder=100)
        pl.pcolormesh(pixels_red_hr, np.arange(46, 85.01, 1), matrix_resolution_red_hr_theo_slice1, cmap=mpl.cm.rainbow,shading='flat', vmin=res_min, vmax=res_max)
        pl.xlabel('Pixels', size=30)
        pl.ylabel('Orders', size=30)
        pl.xticks(size=20)
        pl.yticks(size=20)
        pl.ylim(0, 85)
        if binx ==8:
            pl.xlim(0, 2295)            
        else:
            pl.xlim(0, 4590)
        cbar_ax = fig.add_axes([0.93, 0.11, 0.014, 0.873])
        axcb = pl.colorbar(cax=cbar_ax, format='%0.0f')
        axcb.ax.tick_params(labelsize=20)
        axcb.set_label('Resolution', size=30)
        pl.tight_layout(w_pad=0.1, h_pad=2, rect=(0.0005, 0.0001, 0.935, 0.999))
        pl.savefig('Output/Resolution_map/ESPRESSO/Map_resolution_' + ins_mode + '_' + str(binx) + '_' + str(int(mjd)) + '.png', dpi=100, format='png', bbox_inches='tight')
    return

def Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A, function_map_resolution):
    res_min=120000
    res_max=240000

    header = THAR_LINE_TABLE_THAR_FP_A[0].header

    Time = header['MJD-OBS']

    qual = THAR_LINE_TABLE_THAR_FP_A[1].data['qc']
    ords = THAR_LINE_TABLE_THAR_FP_A[1].data['order'][np.where(qual != 0)]
    pxs = THAR_LINE_TABLE_THAR_FP_A[1].data['x0'][np.where(qual != 0)]
    res = THAR_LINE_TABLE_THAR_FP_A[1].data['resolution'][np.where(qual != 0)]

    if header['HIERARCH ESO INS MODE'] == 'MULTIMR':
        shape_s2d=shape_s2d_ESPRESSO(time_science=Time,ins_mode = header['HIERARCH ESO INS MODE'],bin_x=header['HIERARCH ESO DET BINX'])
        
        orders_blue_hr_slice1 = np.arange(1, 45.01, 1, dtype=int)
        pixels_blue_hr = shape_s2d[0]

        orders_red_hr_slice1 = np.arange(46, 85.01, 1, dtype=int)
        pixels_red_hr = shape_s2d[1]

        matrix_resolution_blue_hr_theo_slice1 = np.zeros((len(np.arange(1, 45.01, 1)), len(pixels_blue_hr)))
        matrix_resolution_red_hr_theo_slice1 = np.zeros((len(np.arange(46, 85.01, 1)), len(pixels_red_hr)))

        orders_blue_slice1 = ords[(ords <= 45)]
        pixels_blue_slice1 = pxs[(ords <= 45)]
        resolution_blue_slice1 = res[(ords <= 45)]

        orders_red_slice1 = ords[(ords > 45)]
        pixels_red_slice1 = pxs[(ords > 45)]
        resolution_red_slice1 = res[(ords > 45)]

        gmodel = Model(function_map_resolution)
        gmodel.set_param_hint('a', value=1., vary=True)
        gmodel.set_param_hint('b', value=1., vary=True)
        gmodel.set_param_hint('c', value=1., vary=True)
        gmodel.set_param_hint('d', value=1., vary=True)
        gmodel.set_param_hint('e', value=1., vary=True)
        gmodel.set_param_hint('f', value=1., vary=True)
        gmodel.set_param_hint('g', value=np.median(resolution_blue_slice1), vary=True)
        pars = gmodel.make_params()
        result_poly_reg_2d_blue_slice1 = gmodel.fit(resolution_blue_slice1,xy=np.array([pixels_blue_slice1, orders_blue_slice1]), params=pars)

        gmodel.set_param_hint('g', value=np.median(resolution_red_slice1), vary=True)
        pars = gmodel.make_params()
        result_poly_reg_2d_red_slice1 = gmodel.fit(resolution_red_slice1, xy=np.array([pixels_red_slice1, orders_red_slice1]), params=pars)

        for i in orders_blue_hr_slice1:
            matrix_resolution_blue_hr_theo_slice1[i - 1, :] = function_map_resolution(np.array([pixels_blue_hr, i-1]),result_poly_reg_2d_blue_slice1.values['a'],result_poly_reg_2d_blue_slice1.values['b'],result_poly_reg_2d_blue_slice1.values['c'],result_poly_reg_2d_blue_slice1.values['d'],result_poly_reg_2d_blue_slice1.values['e'],result_poly_reg_2d_blue_slice1.values['f'],result_poly_reg_2d_blue_slice1.values['g'])

        for i in orders_red_hr_slice1:
            matrix_resolution_red_hr_theo_slice1[i - 46, :] = function_map_resolution(np.array([pixels_red_hr, i-1]),result_poly_reg_2d_red_slice1.values['a'],result_poly_reg_2d_red_slice1.values['b'],result_poly_reg_2d_red_slice1.values['c'],result_poly_reg_2d_red_slice1.values['d'],result_poly_reg_2d_red_slice1.values['e'],result_poly_reg_2d_red_slice1.values['f'],result_poly_reg_2d_red_slice1.values['g'])

        matrix_resolution_blue_hr_theo_slice1 = np.where(matrix_resolution_blue_hr_theo_slice1 == 0., np.nan,matrix_resolution_blue_hr_theo_slice1)
        matrix_resolution_red_hr_theo_slice1 = np.where(matrix_resolution_red_hr_theo_slice1 == 0., np.nan,matrix_resolution_red_hr_theo_slice1)

        plot_resolution_map_MR(pixels_blue_slice1,orders_blue_slice1,resolution_blue_slice1,pixels_blue_hr,matrix_resolution_blue_hr_theo_slice1,
                               pixels_red_slice1,orders_red_slice1,resolution_red_slice1,pixels_red_hr,matrix_resolution_red_hr_theo_slice1,
                               res_min,res_max,header['HIERARCH ESO INS MODE'],header['HIERARCH ESO DET BINX'],header['MJD-OBS'],
                               Doit=True)

        resolution_map = np.zeros((85, len(shape_s2d[1])))

        resolution_map[:45, int(pixels_blue_hr[0]):int(pixels_blue_hr[-1]) + 1] = matrix_resolution_blue_hr_theo_slice1
        resolution_map[45:, :] = matrix_resolution_red_hr_theo_slice1
        resolution_map = np.where(resolution_map == 0., np.nan, resolution_map)

    else:
        shape_s2d=shape_s2d_ESPRESSO(time_science=Time,ins_mode = header['HIERARCH ESO INS MODE'],bin_x=header['HIERARCH ESO DET BINX'])
        orders_blue_hr_slice1 = np.arange(1, 90.01, 1, dtype=int)[::2]
        orders_blue_hr_slice2 = np.arange(1, 90.01, 1, dtype=int)[1::2]
        pixels_blue_hr = shape_s2d[0]

        orders_red_hr_slice1 = np.arange(91, 170.01, 1, dtype=int)[::2]
        orders_red_hr_slice2 = np.arange(91, 170.01, 1, dtype=int)[1::2]
        pixels_red_hr = shape_s2d[1]

        matrix_resolution_blue_hr_theo_slice1 = np.zeros((len(np.arange(1, 90.01, 1)), len(pixels_blue_hr)))
        matrix_resolution_blue_hr_theo_slice2 = np.zeros((len(np.arange(1, 90.01, 1)), len(pixels_blue_hr)))
        matrix_resolution_red_hr_theo_slice1 = np.zeros((len(np.arange(91, 170.01, 1)), len(pixels_red_hr)))
        matrix_resolution_red_hr_theo_slice2 = np.zeros((len(np.arange(91, 170.01, 1)), len(pixels_red_hr)))

        orders_blue_slice1 = ords[(ords <= 90) & (ords % 2 != 0)]
        pixels_blue_slice1 = pxs[(ords <= 90) & (ords % 2 != 0)]
        resolution_blue_slice1 = res[(ords <= 90) & (ords % 2 != 0)]

        orders_blue_slice2 = ords[(ords <= 90) & (ords % 2 == 0)]
        pixels_blue_slice2 = pxs[(ords <= 90) & (ords % 2 == 0)]
        resolution_blue_slice2 = res[(ords <= 90) & (ords % 2 == 0)]

        orders_red_slice1 = ords[(ords > 90) & (ords % 2 != 0)]
        pixels_red_slice1 = pxs[(ords > 90) & (ords % 2 != 0)]
        resolution_red_slice1 = res[(ords > 90) & (ords % 2 != 0)]

        orders_red_slice2 = ords[(ords > 90) & (ords % 2 == 0)]
        pixels_red_slice2 = pxs[(ords > 90) & (ords % 2 == 0)]
        resolution_red_slice2 = res[(ords > 90) & (ords % 2 == 0)]

        gmodel = Model(function_map_resolution)
        gmodel.set_param_hint('a', value=1., vary=True)
        gmodel.set_param_hint('b', value=1., vary=True)
        gmodel.set_param_hint('c', value=1., vary=True)
        gmodel.set_param_hint('d', value=1., vary=True)
        gmodel.set_param_hint('e', value=1., vary=True)
        gmodel.set_param_hint('f', value=1., vary=True)
        gmodel.set_param_hint('g', value=np.median(resolution_blue_slice1), vary=True)
        pars = gmodel.make_params()
        result_poly_reg_2d_blue_slice1 = gmodel.fit(resolution_blue_slice1,xy=np.array([pixels_blue_slice1, orders_blue_slice1]), params=pars)

        gmodel.set_param_hint('g', value=np.median(resolution_blue_slice2), vary=True)
        pars = gmodel.make_params()
        result_poly_reg_2d_blue_slice2 = gmodel.fit(resolution_blue_slice2,xy=np.array([pixels_blue_slice2, orders_blue_slice2]), params=pars)

        gmodel.set_param_hint('g', value=np.median(resolution_red_slice1), vary=True)
        pars = gmodel.make_params()
        result_poly_reg_2d_red_slice1 = gmodel.fit(resolution_red_slice1,xy=np.array([pixels_red_slice1, orders_red_slice1]), params=pars)

        gmodel.set_param_hint('g', value=np.median(resolution_red_slice2), vary=True)
        pars = gmodel.make_params()
        result_poly_reg_2d_red_slice2 = gmodel.fit(resolution_red_slice2,xy=np.array([pixels_red_slice2, orders_red_slice2]), params=pars)

        for i in orders_blue_hr_slice1:
            matrix_resolution_blue_hr_theo_slice1[i - 1, :] = function_map_resolution(np.array([pixels_blue_hr, i-1]),result_poly_reg_2d_blue_slice1.values['a'],result_poly_reg_2d_blue_slice1.values['b'],result_poly_reg_2d_blue_slice1.values['c'],result_poly_reg_2d_blue_slice1.values['d'],result_poly_reg_2d_blue_slice1.values['e'],result_poly_reg_2d_blue_slice1.values['f'],result_poly_reg_2d_blue_slice1.values['g'])

        for i in orders_blue_hr_slice2:
            matrix_resolution_blue_hr_theo_slice2[i - 1, :] = function_map_resolution(np.array([pixels_blue_hr, i-1]),result_poly_reg_2d_blue_slice2.values['a'],result_poly_reg_2d_blue_slice2.values['b'],result_poly_reg_2d_blue_slice2.values['c'],result_poly_reg_2d_blue_slice2.values['d'],result_poly_reg_2d_blue_slice2.values['e'],result_poly_reg_2d_blue_slice2.values['f'],result_poly_reg_2d_blue_slice2.values['g'])

        for i in orders_red_hr_slice1:
            matrix_resolution_red_hr_theo_slice1[i - 91, :] = function_map_resolution(np.array([pixels_red_hr, i-1]),result_poly_reg_2d_red_slice1.values['a'],result_poly_reg_2d_red_slice1.values['b'],result_poly_reg_2d_red_slice1.values['c'],result_poly_reg_2d_red_slice1.values['d'],result_poly_reg_2d_red_slice1.values['e'],result_poly_reg_2d_red_slice1.values['f'],result_poly_reg_2d_red_slice1.values['g'])

        for i in orders_red_hr_slice2:
            matrix_resolution_red_hr_theo_slice2[i - 91, :] = function_map_resolution(np.array([pixels_red_hr, i-1]),result_poly_reg_2d_red_slice2.values['a'],result_poly_reg_2d_red_slice2.values['b'],result_poly_reg_2d_red_slice2.values['c'],result_poly_reg_2d_red_slice2.values['d'],result_poly_reg_2d_red_slice2.values['e'],result_poly_reg_2d_red_slice2.values['f'],result_poly_reg_2d_red_slice2.values['g'])

        matrix_resolution_blue_hr_theo_slice1 = np.where(matrix_resolution_blue_hr_theo_slice1 == 0., np.nan,matrix_resolution_blue_hr_theo_slice1)
        matrix_resolution_blue_hr_theo_slice2 = np.where(matrix_resolution_blue_hr_theo_slice2 == 0., np.nan,matrix_resolution_blue_hr_theo_slice2)
        matrix_resolution_red_hr_theo_slice1 = np.where(matrix_resolution_red_hr_theo_slice1 == 0., np.nan,matrix_resolution_red_hr_theo_slice1)
        matrix_resolution_red_hr_theo_slice2 = np.where(matrix_resolution_red_hr_theo_slice2 == 0., np.nan,matrix_resolution_red_hr_theo_slice2)

        plot_resolution_map_slice_UHR_HR(pixels_blue_slice1,orders_blue_slice1,resolution_blue_slice1,pixels_blue_hr,matrix_resolution_blue_hr_theo_slice1,pixels_red_slice1,
                                         orders_red_slice1,resolution_red_slice1,pixels_red_hr,matrix_resolution_red_hr_theo_slice1,res_min,res_max,
                                         pixels_blue_slice2,orders_blue_slice2,resolution_blue_slice2,matrix_resolution_blue_hr_theo_slice2,pixels_red_slice2,
                                         orders_red_slice2,resolution_red_slice2,matrix_resolution_red_hr_theo_slice2,header['HIERARCH ESO INS MODE'],header['HIERARCH ESO DET BINX'],header['MJD-OBS'],
                                         Doit=True)
        plot_resolution_map_UHR_HR(pixels_blue_slice1,orders_blue_slice1,resolution_blue_slice1,pixels_blue_hr,matrix_resolution_blue_hr_theo_slice1,pixels_red_slice1,
                                         orders_red_slice1,resolution_red_slice1,pixels_red_hr,matrix_resolution_red_hr_theo_slice1,res_min,res_max,
                                         pixels_blue_slice2,orders_blue_slice2,resolution_blue_slice2,matrix_resolution_blue_hr_theo_slice2,pixels_red_slice2,
                                         orders_red_slice2,resolution_red_slice2,matrix_resolution_red_hr_theo_slice2,header['HIERARCH ESO INS MODE'],header['HIERARCH ESO DET BINX'],header['MJD-OBS'],
                                         Doit=True)

        resolution_map = np.zeros((170, len(shape_s2d[1])))

        resolution_map[:90, int(pixels_blue_hr[0]):int(pixels_blue_hr[-1]) + 1] = np.where(
            np.isnan(matrix_resolution_blue_hr_theo_slice1), matrix_resolution_blue_hr_theo_slice2,
            matrix_resolution_blue_hr_theo_slice1)
        resolution_map[90:, :] = np.where(np.isnan(matrix_resolution_red_hr_theo_slice1),
                                          matrix_resolution_red_hr_theo_slice2, matrix_resolution_red_hr_theo_slice1)
        resolution_map = np.where(resolution_map == 0., np.nan, resolution_map)
    prihdu = fits.PrimaryHDU(header=THAR_LINE_TABLE_THAR_FP_A[0].header)
    prihdu.writeto('Static_resolution/ESPRESSO/r.' + header['ARCFILE'][:-5] + '_RESOLUTION_MAP.fits', overwrite=True)
    fits.append('Static_resolution/ESPRESSO/r.' + header['ARCFILE'][:-5] + '_RESOLUTION_MAP.fits', resolution_map)
    return




########################
### files to compute ###
########################



# #UHR
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/UHR/r.ESPRE.2018-09-09T12:18:49.369_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/UHR/r.ESPRE.2019-04-06T11:52:27.477_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/UHR/r.ESPRE.2019-11-06T11:06:36.913_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
#file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/UHR/r.ESPRE.2019-11-26T15:27:21.138_THAR_LINE_TABLE_FP_THAR_B.fits') # fiber B
#Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# #HR 1x1
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/HR/1x1/r.ESPRE.2018-07-04T21:28:53.759_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/HR/1x1/r.ESPRE.2019-01-05T19:53:55.501_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/HR/1x1/r.ESPRE.2019-11-19T10:10:12.384_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)

# #HR 2x1
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/HR/2x1/r.ESPRE.2018-09-05T14:01:58.063_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/HR/2x1/r.ESPRE.2019-03-30T21:28:32.060_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/HR/2x1/r.ESPRE.2019-09-26T11:06:01.271_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)

# #MR4x2
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/4x2/r.ESPRE.2018-07-08T14:51:53.873_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/4x2/r.ESPRE.2018-12-01T11:25:27.377_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/4x2/r.ESPRE.2019-11-05T12:23:45.139_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)

# #MR8x4
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/8x4/r.ESPRE.2018-07-06T11:48:22.862_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/8x4/r.ESPRE.2018-10-24T14:16:52.394_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)
# file_to_map_resolution = fits.open('Input_files_statics/Resolution_table/ESPRESSO/8x4/r.ESPRE.2019-11-05T12:59:57.138_THAR_LINE_TABLE_THAR_FP_A.fits')
# Instrumental_resolution(THAR_LINE_TABLE_THAR_FP_A=file_to_map_resolution,function_map_resolution=polynomial_regression_2d)


"""



static_hitran_file('Input_files_statics/Molecules/CH4/NIRPS/12C_H4__9000_18000.txt',
                    'Input_files_statics/Molecules/CH4/NIRPS/QT.txt','CH4',
                    S_cut=10 ** -24,instrument='NIRPS')

# static_hitran_file('Input_files_statics/Molecules/CO/NIRPS/12C_16O__9000_18000.txt',
#                     'Input_files_statics/Molecules/CO/NIRPS/QT.txt','CO',
#                     S_cut=10 ** -26,instrument='NIRPS')

static_hitran_file('Input_files_statics/Molecules/CO2/NIRPS/12C_16O2__9000_18000.txt',
                    'Input_files_statics/Molecules/CO2/NIRPS/QT.txt','CO2',
                    S_cut=10 ** -26,instrument='NIRPS')

static_hitran_file('Input_files_statics/Molecules/H2O/NIRPS/H2_16O__9000_18000.txt',
                    'Input_files_statics/Molecules/H2O/NIRPS/QT.txt','H2O',
                    S_cut=10 ** -26,instrument='NIRPS')

static_hitran_file('Input_files_statics/Molecules/O2/NIRPS/16O2__9000_18000.txt',
                    'Input_files_statics/Molecules/O2/NIRPS/QT.txt','O2',
                    S_cut=10 ** -28,instrument='NIRPS')


lines_to_fit('Static_model/NIRPS/Static_hitran_qt_H2O.fits',
             'H2O',wavenumber_born=np.array([10**8/12850,10**8/15170]),number_of_lines=50,
             region_excluded=np.array([10**8/13250,10**8/15000]),instrument='NIRPS')

lines_to_fit('Static_model/NIRPS/Static_hitran_qt_O2.fits',
             'O2',wavenumber_born=np.array([10**8/12450,10**8/12950]),number_of_lines=30,
             region_excluded=np.array([10**8/12955,10**8/12960]),instrument='NIRPS')

lines_to_fit('Static_model/NIRPS/Static_hitran_qt_CH4.fits',
             'CH4',wavenumber_born=np.array([10**8/16200,10**8/16900]),number_of_lines=16,instrument='NIRPS')

lines_to_fit('Static_model/NIRPS/Static_hitran_qt_CO2.fits',
             'CO2',wavenumber_born=np.array([10**8/15150,10**8/16200]),number_of_lines=50,
             region_excluded=np.array([10**8/12680,10**8/12710]),instrument='NIRPS')

"""
"""

static_hitran_file('Input_files_statics/Molecules/H2O/ESPRESSO/H2_16O__3700_8000.txt',
                  'Input_files_statics/Molecules/H2O/ESPRESSO/QT.txt','H2O',
                  S_cut=10 ** -26,instrument='ESPRESSO')

static_hitran_file('Input_files_statics/Molecules/O2/ESPRESSO/16O2__3700_8000.txt',
                  'Input_files_statics/Molecules/O2/ESPRESSO/QT.txt','O2',
                  S_cut=10 ** -28,instrument='ESPRESSO')


lines_to_fit('Static_model/ESPRESSO/Static_hitran_qt_H2O.fits',
             'H2O',wavenumber_born=np.array([10**8/7150,10**8/7350]),number_of_lines=20,
             lines_excluded=np.array([10**8/7186.51031647,10**8/7208.41193416,10**8/7236.72396574]),instrument='ESPRESSO')

lines_to_fit('Static_model/ESPRESSO/Static_hitran_qt_O2.fits',
             'O2',wavenumber_born=np.array([10**8/6000,10**8/6400]),number_of_lines=12,
             region_excluded=np.array([10**8/6855,10**8/6918]),instrument='ESPRESSO')

"""


#(TAVELLA)

#HARPS_N

#H2O
static_hitran_file('Input_files_statics/Molecules/H2O/HARPS_N/H2_16O__3750_6950.txt',
                  'Input_files_statics/Molecules/H2O/HARPS_N/QT.txt','H2O',
                  S_cut=10 ** -26,instrument='HARPS_N')

lines_to_fit('Static_model/HARPS_N/Static_hitran_qt_H2O.fits',
             'H2O',wavenumber_born=np.array([10**8/6450,10**8/6600]),number_of_lines=20, instrument='HARPS_N')

#O2
static_hitran_file('Input_files_statics/Molecules/O2/HARPS_N/16O2__3750_6950.txt',
                  'Input_files_statics/Molecules/O2/HARPS_N/QT.txt','O2',
                  S_cut=10 ** -28,instrument='HARPS_N')

lines_to_fit('Static_model/HARPS_N/Static_hitran_qt_O2.fits',
             'O2',wavenumber_born=np.array([10**8/6250,10**8/6350]),number_of_lines=12,
             region_excluded=np.array([10**8/6855,10**8/6918]),instrument='HARPS_N')

#HARPS

#H2O
static_hitran_file('Input_files_statics/Molecules/H2O/HARPS/H2_16O__3750_6950.txt',
                  'Input_files_statics/Molecules/H2O/HARPS/QT.txt','H2O',
                  S_cut=10 ** -26,instrument='HARPS')

lines_to_fit('Static_model/HARPS/Static_hitran_qt_H2O.fits',
             'H2O',wavenumber_born=np.array([10**8/6450,10**8/6600]),number_of_lines=20, instrument='HARPS')

#O2
static_hitran_file('Input_files_statics/Molecules/O2/HARPS/16O2__3750_6950.txt',
                  'Input_files_statics/Molecules/O2/HARPS/QT.txt','O2',
                  S_cut=10 ** -28,instrument='HARPS')

lines_to_fit('Static_model/HARPS/Static_hitran_qt_O2.fits',
             'O2',wavenumber_born=np.array([10**8/6250,10**8/6350]),number_of_lines=12,
             region_excluded=np.array([10**8/6855,10**8/6918]),instrument='HARPS')

#CARMENES_VIS

#H2O
static_hitran_file('Input_files_statics/Molecules/H2O/CARMENES_VIS/H2_16O__5000_9800.txt',
                  'Input_files_statics/Molecules/H2O/CARMENES_VIS/QT.txt','H2O',
                  S_cut=10 ** -26,instrument='CARMENES_VIS')

lines_to_fit('Static_model/CARMENES_VIS/Static_hitran_qt_H2O.fits',
             'H2O',wavenumber_born=np.array([10**8/7150,10**8/7350]),number_of_lines=20,
             lines_excluded=np.array([10**8/7186.51031647,10**8/7208.41193416,10**8/7236.72396574]), instrument='CARMENES_VIS')

#O2
static_hitran_file('Input_files_statics/Molecules/O2/CARMENES_VIS/16O2__5000_9800.txt',
                  'Input_files_statics/Molecules/O2/CARMENES_VIS/QT.txt','O2',
                  S_cut=10 ** -28,instrument='CARMENES_VIS') 

lines_to_fit('Static_model/CARMENES_VIS/Static_hitran_qt_O2.fits',
             'O2',wavenumber_born=np.array([10**8/6250,10**8/6350]),number_of_lines=12,
             region_excluded=np.array([10**8/6855,10**8/6918]),instrument='CARMENES_VIS')

#different options for restricted fit range + excluded regions
#lines_to_fit('Static_model/CARMENS_VIS/Static_hitran_qt_O2.fits',
            # 'O2',wavenumber_born=np.array([10**8/6250,10**8/6350]),number_of_lines=12, # np.array([10**8/6920,10**8/7000]) or np.array([10**8/7690,10**8/7750])
            # region_excluded=np.array([10**8/6855,10**8/6918], [10**8/7590,10**8/7690]),instrument='CARMENS_VIS')



"""

lines_to_fit('Static_model/HARPS_N/Static_hitran_qt_H2O.fits',
             'H2O',wavenumber_born=np.array([10**8/7150,10**8/7350]),number_of_lines=20,
             lines_excluded=np.array([10**8/7186.51031647,10**8/7208.41193416,10**8/7236.72396574]),instrument='HARPS_N')

lines_to_fit('Static_model/HARPS_N/Static_hitran_qt_O2.fits',
             'O2',wavenumber_born=np.array([10**8/6250,10**8/6350]),number_of_lines=12,
             region_excluded=np.array([10**8/6855,10**8/6918]),instrument='HARPS_N')






"""














#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from copy import deepcopy
from scipy.interpolate import interp1d
import pandas as pd
from pathos.multiprocessing import Pool
from scipy.signal import savgol_filter
from os import makedirs
from os.path import exists as path_exist
from ..ANTARESS_analysis.ANTARESS_inst_resp import calc_FWHM_inst
from ..ANTARESS_general.utils import init_parallel_func,np_where1D,is_odd,stop,dataload_npz,datasave_npz,check_data


def process_spectral_cont(vis_mode,data_type_gen,inst,data_dic,gen_dic,vis):
    r"""**Stellar continuum: main routine.**
    
    Returns generic stellar continuum from binned master spectrum.

    Args:
        TBD
    
    Returns:
        TBD
    
    """       
    print('   > Defining stellar continuum on '+gen_dic['type_name'][data_type_gen]+' master') 
    data_inst = data_dic[inst]   

    #Using master from several visits
    if vis_mode=='multivis':
        vis_det='binned'
        if data_inst['type']!='spec1D':stop('Spectra must be 1D') 

    #Using master from single visit
    elif vis_mode=='':
        vis_det=vis   
        if data_inst[vis]['type']!='spec1D':stop('Spectra must be 1D')    

    #Processing data    
    save_data_paths = gen_dic['save_data_dir']+'Stellar_cont_'+data_type_gen+'/'+inst+'_'+vis_det+'/'
    if not path_exist(save_data_paths):makedirs(save_data_paths)  
    if (gen_dic['calc_'+data_type_gen+'_stcont']):
        print('         Calculating data')
        prop_dic = deepcopy(data_dic[data_type_gen]) 
    
        #Retrieve binning information
        data_bin = dataload_npz(gen_dic['save_data_dir']+data_type_gen+'bin_data/'+gen_dic['add_txt_path'][data_type_gen]+inst+'_'+vis_det+'_'+prop_dic['dim_bin']+'_add')
    
        #Retrieve master spectrum
        if data_bin['n_exp']>1:stop('Bin data into a single master spectrum')
        data_mast = dataload_npz(gen_dic['save_data_dir']+data_type_gen+'bin_data/'+gen_dic['add_txt_path'][data_type_gen]+inst+'_'+vis_det+'_'+prop_dic['dim_bin']+str(0))

        #Check for alignment
        if (not gen_dic['align_'+data_type_gen]) or ((data_type_gen=='DI') and (not gen_dic['align_DI'] and data_bin['sysvel']==0.)):
            stop('Data must have been aligned in the stellar rest frame')
            
        #Check for in-transit contamination
        if (data_type_gen=='DI') and data_bin['in_inbin']:
            stop('Disk-integrated master contain in-transit profiles')    
    
        #Limit master to minimum definition range
        idx_def_mast = np_where1D(data_mast['cond_def'][0])
        flux_mast = data_mast['flux'][:,idx_def_mast[0]:idx_def_mast[-1]+1]
        cen_bins_mast = data_mast['cen_bins'][:,idx_def_mast[0]:idx_def_mast[-1]+1]
        cond_def_mast = data_mast['cond_def'][:,idx_def_mast[0]:idx_def_mast[-1]+1]
        edge_bins_mast = data_mast['edge_bins'][:,idx_def_mast[0]:idx_def_mast[-1]+2]
        
        #Stellar continuum
        min_edge_ord = cen_bins_mast[0,0]
        dic_sav = {}
        _,cont_func_dic,_ = calc_spectral_cont(1,[0],flux_mast,cen_bins_mast,edge_bins_mast,cond_def_mast,None,None,inst,gen_dic['contin_roll_win'][inst],gen_dic['contin_smooth_win'][inst],gen_dic['contin_locmax_win'][inst],\
                                                         gen_dic['contin_stretch'][inst],gen_dic['contin_pinR'][inst],min_edge_ord,dic_sav,1)
            
        #Saving data     
        datasave_npz(save_data_paths+'St_cont',{'cont_func_dic': cont_func_dic[0]})

    else:
        data_paths={'path':save_data_paths+'St_cont'}
        check_data(data_paths)          
        
    return None



def calc_spectral_cont(nord,iord_proc_list,flux_mast_in,cen_bins_mast,edge_bins_mast,cond_def_mast,flux_Earth_all_in,cond_def_all_in,inst,roll_win,smooth_win,locmax_win,par_stretch,\
                       contin_pinR,min_edge_ord,dic_sav,nthreads):
    r"""**Stellar continuum: wrap-up for calculation.**
    
    Calls direct calculation or multi-threaded one.

    Args:
        TBD
    
    Returns:
        TBD
    
    """     
    mean_flux_mast = np.zeros(nord,dtype=float)*np.nan
    if flux_mast_in is None:
        flux_mast=np.repeat(None,len(iord_proc_list))
        flux_Earth_all = deepcopy(flux_Earth_all_in)   
        cond_def_all = deepcopy(cond_def_all_in)    
    else:
        flux_mast = deepcopy(flux_mast_in)
        flux_Earth_all=np.tile(None,[1,len(iord_proc_list)])
        cond_def_all=np.tile(None,[1,len(iord_proc_list)])
    common_args = (inst,roll_win,smooth_win,locmax_win,par_stretch,contin_pinR,min_edge_ord)
    if nthreads>1:mean_flux_mast[iord_proc_list],dic_sav_loc = para_sub_spectral_cont(sub_spectral_cont,nthreads,len(iord_proc_list),[iord_proc_list,cond_def_mast,cen_bins_mast,flux_mast,edge_bins_mast,flux_Earth_all,cond_def_all],common_args)                           
    else:mean_flux_mast[iord_proc_list],dic_sav_loc = sub_spectral_cont(iord_proc_list,cond_def_mast,cen_bins_mast,flux_mast,edge_bins_mast,flux_Earth_all,cond_def_all,*common_args)  
    dic_sav.update(dic_sav_loc)

    #----------------------------------------------------
    #Define continuum function for current order
    #    - performed outside of the parallelized loop because for some reason it raises issues with 'cont_func_dic'
    cont_func_dic={}
    for isub_ord,iord in enumerate(iord_proc_list):
        cont_func_dic[iord] = interp1d(dic_sav[iord]['wav_max'],dic_sav[iord]['flux_max'],fill_value='extrapolate')  

    return mean_flux_mast,cont_func_dic,dic_sav



def sub_spectral_cont(iord_proc_list,cond_def_mast,cen_bins_mast,flux_mast,edge_bins_mast,flux_Earth_all,cond_def_all,inst,roll_win,smooth_win,locmax_win,par_stretch,contin_pinR,min_edge_ord):
    r"""**Stellar continuum: direct calculation.**
    
    Estimates stellar continuum from master spectrum.

    Args:
        TBD
    
    Returns:
        TBD
    
    """  
    mean_flux_mast = np.zeros(len(iord_proc_list),dtype=float)*np.nan
    dic_sav={}
    low_edge_bins = edge_bins_mast[:,0:-1]
    high_edge_bins = edge_bins_mast[:,1::]
    dcen_bins = high_edge_bins - low_edge_bins  
    for isub_ord,iord in enumerate(iord_proc_list):

        #Master tables over order
        cond_def = cond_def_mast[isub_ord]
        cen_bins_ord = cen_bins_mast[isub_ord,cond_def]
        low_edge_bins_ord = low_edge_bins[isub_ord][cond_def]
        high_edge_bins_ord = high_edge_bins[isub_ord][cond_def]
        dcen_bins_ord = dcen_bins[isub_ord][cond_def]
        
        #Calculate master from median flux if undefined
        if flux_mast[isub_ord] is None:
            flux_mast_ord = np.zeros(np.sum(cond_def))*np.nan                 
            for ipix_sub,ipix in enumerate(np_where1D(cond_def)):flux_mast_ord[ipix_sub] = np.median(flux_Earth_all[cond_def_all[:,isub_ord,ipix],isub_ord,ipix])
        else:
            flux_mast_ord = flux_mast[isub_ord,cond_def]
       
        #Mean master flux in order
        mean_flux_mast[isub_ord] = np.sum(flux_mast_ord*dcen_bins_ord)/np.sum(dcen_bins_ord)

        #Mean bin size over order (A)                
        dwav_mean = np.mean(dcen_bins_ord)

        #-------------------------------------------------------
        #Preliminary peak exclusion
        
        #Half-integer range over which peaks are excluded
        #    - exclusion range set to 3*fwhm of the inst
        fwhm_wav = calc_FWHM_inst(inst,np.mean(cen_bins_ord))  
        hn_exc = int(np.round(0.5*(3*fwhm_wav /dwav_mean)))

        #Apply successive sigma-clipping to remove spurious peaks
        #    - this step should not be necessary if the cosmic correction module was applied
        pd_spectre = pd.DataFrame(flux_mast_ord)
        n_roll = int(np.round(roll_win/dwav_mean))
        for iteration in range(5): 
            
            #Envelopes of the spectrum at different quantile
            maxi_roll_fast = np.ravel(pd_spectre.rolling(5*n_roll,min_periods=1,center=True).quantile(0.99))
            Q3_fast = np.ravel(pd_spectre.rolling(n_roll,min_periods=1,center=True).quantile(0.75)) 
            median = np.ravel(pd_spectre.rolling(n_roll,min_periods=1,center=True).quantile(0.50))

            #Condition of peak removal
            sup_fast = Q3_fast+3*(Q3_fast-median)  
            mask = (flux_mast_ord>sup_fast) & (flux_mast_ord>maxi_roll_fast)
            
            #Replace peaks and the chosen range around them
            mask_range = deepcopy(mask)
            for j in range(1,hn_exc):mask_range  |= np.roll(mask,-j) | np.roll(mask,j) 
            flux_mast_ord[mask_range] = median[mask_range]

        #-------------------------------------------------------
        #Smoothing
        #    - using Savitzky-Golay filter   
        n_smooth_win = int(smooth_win/dwav_mean)
        if not is_odd(n_smooth_win):n_smooth_win+=1
        smooth_flux = savgol_filter(flux_mast_ord, n_smooth_win, 3)

        #Suppress the smoothing of sharp peaks that create sinc-like wiggle
        spectre_backup = flux_mast_ord.copy()
        median = np.median(abs(spectre_backup-smooth_flux))
        IQ = np.percentile(abs(spectre_backup-smooth_flux),75) - median
        mask_out = np.where(abs(spectre_backup-smooth_flux)>(median+20*IQ))[0]
        mask_out = np.unique(mask_out+np.arange(-n_smooth_win,n_smooth_win+1,1)[:,np.newaxis])
        mask_out = mask_out[(mask_out>=0)&(mask_out<len(cen_bins_ord))].astype('int')
        smooth_flux[mask_out] = spectre_backup[mask_out] 

        #Set negative flux values to 0
        smooth_flux[smooth_flux<0] = 0. 

        #-------------------------------------------------------
        #Identification of local maxima 
        hn_locmax = int(np.round(0.5*locmax_win/dwav_mean))
        smooth_flux_noedge = smooth_flux[hn_locmax:-hn_locmax]
        maxima = np.ones(len(smooth_flux_noedge))
        for k in range(1,hn_locmax):
            maxima *= 0.5*(1+np.sign(smooth_flux_noedge - smooth_flux[hn_locmax-k:-hn_locmax-k]))*0.5*(1+np.sign(smooth_flux_noedge - smooth_flux[hn_locmax+k:-hn_locmax+k]))
        index = np_where1D(maxima==1)+hn_locmax
        maxima_flux = smooth_flux[index]
        maxima_cen_bins = cen_bins_ord[index]

        #Dealing with edge issues
        if maxima_flux[0] < smooth_flux[0]:
            maxima_cen_bins = np.insert(maxima_cen_bins,0,cen_bins_ord[0])
            maxima_flux = np.insert(maxima_flux,0,smooth_flux[0])
        if maxima_flux[-1] < smooth_flux[-1]:
            maxima_cen_bins = np.hstack([maxima_cen_bins,cen_bins_ord[-1]])
            maxima_flux = np.hstack([maxima_flux,smooth_flux[-1]])

        #Removing cosmics
        pd_max_flux = pd.DataFrame(maxima_flux)
        median = np.ravel(pd_max_flux.rolling(10,center=True).quantile(0.50))
        IQ = np.ravel(pd_max_flux.rolling(10,center=True).quantile(0.75)) - median
        IQ[np.isnan(IQ)] = smooth_flux.max()
        median[np.isnan(median)] = smooth_flux.max()
        cond_nopeaks = (maxima_flux <= median + 20 * IQ)
        maxima_cen_bins = maxima_cen_bins[cond_nopeaks]
        maxima_flux = maxima_flux[cond_nopeaks]

        #-------------------------------------------------------
        #Stretching
        #    - Y-axis stretching value
        #    - to scale the x and y axis    
        normalisation = (np.max(flux_mast_ord) - np.min(flux_mast_ord))/(high_edge_bins_ord[-1]  - low_edge_bins_ord[0])
        maxima_flux = maxima_flux/par_stretch/normalisation

        #----------------------------------------------------
        #Rolling pin
        waves = maxima_cen_bins - maxima_cen_bins[:,np.newaxis]
        distance = np.sign(waves)*np.sqrt((waves)**2+(maxima_flux - maxima_flux[:,np.newaxis])**2)
        distance[distance<0] = 0

        #Chromatic radius, constant in velocity space
        radius = np.repeat(contin_pinR,len(maxima_cen_bins))*maxima_cen_bins/min_edge_ord
        radius[0] = radius[0]/1.5
        
        #Applying rolling pin
        numero = np.arange(len(distance)).astype('int')
        j = 0
        keep = [0]
        while (len(maxima_cen_bins)-j>3): 
            par_R = float(radius[j]) #take the radius from the penality law
            mask = (distance[j,:]>0)&(distance[j,:]<2.*par_R) #recompute the points closer than the diameter if radius changed with the penality      
            while np.sum(mask)==0:
                par_R *=1.5
                mask = (distance[j,:]>0)&(distance[j,:]<2.*par_R) #recompute the points closer than the diameter if radius changed with the penality      
            p1 = np.array([maxima_cen_bins[j],maxima_flux[j]]).T #vector of all the local maxima 
            p2 = np.array([maxima_cen_bins[mask],maxima_flux[mask]]).T #vector of all the maxima in the diameter zone
            delta = p2 - p1 # delta x delta y
            c = np.sqrt(delta[:,0]**2+delta[:,1]**2) # euclidian distance 
            h = np.sqrt(par_R**2-0.25*c**2) 
            cx = p1[0] + 0.5*delta[:,0] - h/c*delta[:,1] #x coordinate of the circles center
            cy = p1[1] + 0.5*delta[:,1] + h/c*delta[:,0] #y coordinates of the circles center
            cond1 = (cy-p1[1])>=0
            thetas = cond1*(-1*np.arccos((cx - p1[0])/par_R)+np.pi) + (1-1*cond1)*(-1*np.arcsin((cy - p1[1])/par_R) + np.pi)
            j2 = thetas.argmin()
            j = numero[mask][j2] #take the numero of the local maxima falling in the diameter zone
            keep.append(j)
        maxima_flux = maxima_flux[keep] #we only keep the local maxima with the rolling pin condition
        maxima_cen_bins = maxima_cen_bins[keep]

        #----------------------------------------------------
        #Removing outlying maxima
        diff_deri = abs(np.diff(np.diff(maxima_flux)/np.diff(maxima_cen_bins)))
        mask_out = diff_deri  > (np.percentile(diff_deri,99.5))
        mask_out = np.array([False]+mask_out.tolist()+[False])
        maxima_cen_bins = maxima_cen_bins[~mask_out]
        maxima_flux = maxima_flux[~mask_out]
        
        #Unstretch maxima spectrum
        norm_maxima_flux = maxima_flux*normalisation*par_stretch

        #Save
        dic_sav[iord]={'wav_max':maxima_cen_bins,'flux_max':norm_maxima_flux}  
        if flux_mast[isub_ord] is None:
            dic_sav[iord].update({'wav_mast':cen_bins_ord,'flux_mast':flux_mast_ord})       

    return mean_flux_mast,dic_sav


def para_sub_spectral_cont(func_input,nthreads,n_elem,y_inputs,common_args):  
    r"""**Multithreading routine for sub_spectral_cont().**

    Args:
        func_input (function): multi-threaded function
        nthreads (int): number of threads
        n_elem (int): number of elements to thread
        y_inputs (list): threadable function inputs 
        common_args (tuple): common function inputs
    
    Returns:
        y_output (None or specific to func_input): function outputs 
    
    """
    pool_proc = Pool(processes=nthreads)   #cannot be passed through lmfit
    ind_chunk_list=init_parallel_func(nthreads,n_elem)
    chunked_args=[tuple(y_inputs[i][ind_chunk[0]:ind_chunk[1]] for i in range(5))+(y_inputs[5][:,ind_chunk[0]:ind_chunk[1]],y_inputs[6][:,ind_chunk[0]:ind_chunk[1]],)+common_args for ind_chunk in ind_chunk_list]	
    all_results=tuple(tab for tab in pool_proc.starmap(func_input,chunked_args))	

    mean_flux_mast = np.concatenate(tuple(all_results[i][0] for i in range(nthreads)),axis=0)
    dic_sav = {}
    for dic_save_proc in tuple(all_results[i][1] for i in range(nthreads)):dic_sav.update(dic_save_proc)
    y_output=(mean_flux_mast,dic_sav) 
    
    pool_proc.close()
    pool_proc.join() 				
    return y_output
















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from itertools import product as it_product
from ..ANTARESS_general.utils import stop,planck
from ..ANTARESS_general.constant_data import G_usi,Msun
from ..ANTARESS_grids.ANTARESS_coord import calc_zLOS_oblate,frameconv_skystar_to_star


def calc_RVrot(x_st_sky,y_st,istar_rad,st_par):
    r"""**Stellar rotational rv**

    Calculates radial velocity of stellar surface element from rotation (in km/s).     
    The absolute and radial velocity depend on stellar latitude in presence of differential rotation.

    .. math::  
       v &= \Omega R_\mathrm{\star} \\
         &= \Omega_\mathrm{eq} (1-\alpha_\mathrm{rot} y_\mathrm{lat}^2 - \beta_\mathrm{rot} y_\mathrm{lat}^4) R_\mathrm{\star} \\           
         &= v_\mathrm{eq} (1-\alpha_\mathrm{rot} y_\mathrm{lat}^2 - \beta_\mathrm{rot} y_\mathrm{lat}^4) 
         
    The velocity vector in the star frame is defined as

    .. math::    
       v_\mathrm{x,star} &=  v \cos(\Phi)   \\
       v_\mathrm{y,star} &= -v \sin(\Phi)   \\
       v_\mathrm{z,star} &=  0 
        
    Where :math:`\Phi` is the angle between the LOS `z` and the surface element in the `zx` plane.
    The velocity vector in the inclined star frame is then
    
    .. math::    
       v_\mathrm{x,sky star} &=  v \cos(\Phi)   \\
       v_\mathrm{y,sky star} &= -v \sin(\Phi) cos(i_\star)   \\
       v_\mathrm{z,sky star} &= -v \sin(\Phi) sin(i_\star) 

    And the radial velocity along the :math:`z_\mathrm{sky star}` axis, defined as negative toward the observer, is then
    
    .. math:: 
       rv &= - v_\mathrm{z,sky star}    \\
          &= v \sin(\Phi) sin(i_\star)    \\
          &= v x_\mathrm{norm} sin(i_\star)    \\          
          &= x_\mathrm{norm} v_\mathrm{eq} sin(i_\star) (1-\alpha_\mathrm{rot} y_\mathrm{lat}^2 - \beta_\mathrm{rot} y_\mathrm{lat}^4)  

    Args:
        TBD
    
    Returns:
        TBD
    
    """   
    Vrot = st_par['veq']*(1.-st_par['alpha_rot']*y_st**2.-st_par['beta_rot']*y_st**4.)
    RVrot = x_st_sky*Vrot*np.sin(istar_rad) 
    return RVrot,Vrot



def calc_CB_RV(ld_coeff,LD_mod,c1_CB,c2_CB,c3_CB,star_params):
    r"""**Stellar convective blueshift**

    Calculates radial velocity of stellar surface element from convective blueshift.

    Args:
        TBD
    
    Returns:
        TBD
    
    """     
    #Convective blueshift contribution
    if (c3_CB != 0.) or (c2_CB != 0.) or (c1_CB != 0.): 
        if star_params['f_GD']>0.:stop('CB not available with GD')

        #---------------------------------------------------------------------        
        #Determination of the constant coefficient for the CB
        #    - after isolating the local stellar profile behind the planet during the transit, we shift them  
        # by the out-of-transit RV obtained from the master-out, unocculted CCFs. This RVout includes not only
        # the systemic RV but also the disk-integrated CB RV (the disk-integrated rotational contribution is null)
        #      thus, the RV we measure and fit here are RVobs = RVobs(st. rot) + RVobs(CB) - RVobs(disk CB)
        #      our model must verify:
        # int( RVmod over disk) = int( RVobs over disk)
        # int( RVmod(st. rot.) + RVmod(poly for CB) over disk) = int( RVobs(st. rot) + RVobs(CB) - RVobs(disk CB), over disk)     
        # int( RVmod(poly for CB) over disk) = int( RVobs(CB) , over disk) - RVobs(disk CB) = 0
        # int( RVmod(poly for CB) over disk) = 0
        # int( phi:0,2pi; theta:0,pi/2 [ I(th) cos(th)*RVconv(th)*Rstar^2*sin(th)*dth*dphi ])/disk_integrated_I = 0
        #        mu = cos(th) and dmu = -sin(th)dth
        # int( mu:0,1 [ I(mu) mu*RVconv(mu)dmu ]) = 0
        #       then for a given limb-darkening law we express c0 as a function of the combination of the other disk integrated coefficients
        #     - with convective blueshift defined as a polynomial in mu:
        # RVconv(mu) = c0 + c1*mu + c2*mu^2 + c3*mu^2 ...
        #       the condition becomes
        # c0*int( mu:0,1 [ I(mu)*mu*dmu ]) + sum(i>=1 , ci*int( mu:0,1 [ I(mu)*mu^(i+1)*dmu ])  = 0    
        # c0 = - sum(i>=1 , ci*Int_i)/Int_0
        #       with Int_i = int( mu:0,1 [ I(mu)*mu^(i+1)*dmu ])
        #            I(mu) = I0*LD(mu) and I0 the specific intensity at disk center disk, which is not involved in the calculations                                 
        # the constant coefficient in our model includes the true constant contribution from the CB and a correction factor due to the RV shift
        if LD_mod == 'uniform' or LD_mod == 'linear' or LD_mod == 'quadratic':
            LD_u1,LD_u2=ld_coeff[0:2]
            Int0 = 0.5 - (1./6.)*LD_u1 - (1./12.)*LD_u2           #int(0:1 , I(mu)*mu*dmu)
            Int1 = (1./3.) - (1./12.)*LD_u1 - (1./30.)*LD_u2      #int(0:1 , I(mu)*mu2*dmu)
            Int2 = (1./4.) - (1./20.)*LD_u1 - (1./60.)*LD_u2      #int(0:1 , I(mu)*mu3*dmu)
            Int3 = (1./5.) - (1./30.)*LD_u1 - (1./105.)*LD_u2     #int(0:1 , I(mu)*mu4*dmu)
        elif LD_mod == 'nonlinear':
            LD_u1,LD_u2,LD_u3,LD_u4=ld_coeff[0:5]
            Int0 = 0.5 - (1./10.)*LD_u1 - (1./6.)*LD_u2 - (3./14.)*LD_u3 - (1./4.)*LD_u4           #int(0:1 , I(mu)*mu*dmu)
            Int1 = (1./3.) - (1./21.)*LD_u1 - (1./12.)*LD_u2 - (1./9.)*LD_u3 - (2./15.)*LD_u4      #int(0:1 , I(mu)*mu2*dmu)
            Int2 = (1./4.) - (1./36.)*LD_u1 - (1./20.)*LD_u2 - (3./44.)*LD_u3 - (1./12.)*LD_u4     #int(0:1 , I(mu)*mu3*dmu)
            Int3 = (1./5.) - (1./55.)*LD_u1 - (1./30.)*LD_u2 - (3./65.)*LD_u3 - (2./35.)*LD_u4     #int(0:1 , I(mu)*mu4*dmu)
        else:stop('LD law undefined')
            
        #Solving the equation int( RVmod(poly for CB) over disk) = 0 we get (in km/s)
        c0_CB = - (c1_CB*Int1 + c2_CB*Int2 +c3_CB*Int3)/Int0

    else:
        c0_CB=0.
    cb_band = np.array([c0_CB,c1_CB,c2_CB,c3_CB])    
    return cb_band




def calc_GD(x_grid_star,y_grid_star,z_grid_star,star_params,gd_band,x_st_sky_grid,y_st_sky_grid):
    r"""**Gravity-darkening intensity**

    Calculates blackbody emission with gravity-darkening, using the formalism from Barnes+2009.
    It accounts from gravity-darkening through temperature variation and local blackbody emission.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    
    #Distance to stellar rotation axis
    rproj_grid_star2 = x_grid_star*x_grid_star + z_grid_star*z_grid_star
    
    #Distance to photosphere
    r_grid_star = np.sqrt(rproj_grid_star2 + y_grid_star*y_grid_star)

    #Local surface gravity vector
    #    - Eq 10 in Barnes+2009
    #    - Vg = gr Vr + grp Vrp  
    # with gr = (-G*Mstar/R^2)   [m3 kg-1 s-2 kg m-2] = [m s-2] 
    #      gr/R = (-G*Mstar/R^3) [s-2]    
    #      grp = om_eq^2*Rp    [rad s-2 m] 
    #      grp/Rp = om_eq^2    [s-2]  
    #    - projected onto the star frame:
    #      Vgx = xstar*((gr/R) + (grp/Rp))
    #      Vgy = ystar*(gr/R) 
    #      Vgz = zstar*((gr/R) + (grp/Rp)) 
    #    - g = sqrt( Vgx^2 + Vgy^2 + Vgz^2 )
    #    - at the pole of the star: 
    # xstar=0,zstar=0,ystar = +- Req*(1-f)
    # R = ystar 
    # gpole = Vgy_pole  
    Rstar_m = star_params['Rstar_km']*1e3                                 
    gr_R = -G_usi*star_params['Mstar']*Msun/(r_grid_star*Rstar_m)**3.
    grp_Rp = star_params['om_eq']**2.
    Vgx_grid_star = x_grid_star*(gr_R + grp_Rp)
    Vgy_grid_star = y_grid_star*gr_R 
    Vgz_grid_star = z_grid_star*(gr_R + grp_Rp) 
    g_grid_star = np.sqrt( Vgx_grid_star**2. + Vgy_grid_star**2. + Vgz_grid_star**2. )

    #Mu angle
    #    - defined as mu = cos(theta), where theta is the angle between the local normal to the photosphere N and the LOS = K
    #    - the cross-product between the local normal and the LOS is vN.vK = N*K*mu
    #    - see calc_zLOS_oblate(), the LOS coordinate of the (visible) photosphere is defined in the sky-projected rest frame via :
    # z = f(x,y) = (-B(y)+sqrt(D(x,y)))/(2*A)
    #   A = (1 - si^2*(1-Rpole^2))
    #   B(y) = 2*y*ci*si*(1 - Rpole^2)
    #   C(x,y) = y^2*si^2*(1- Rpole^2) + Rpole^2*(x^2+ y^2 - 1)
    #   D(x,y) = B(y)^2 - 4*A*C(x,y)
    #      the normal to the photosphere is defined as  
    # N = (-df/dx,-df/dy,1)
    #   df/dx = d[ (-B(y)+sqrt(D(x,y)))/(2*A) ]/dx
    #         = (1/(2*A)) * d[ sqrt(D(x,y)) ]/dx    
    #         =  (1/(2*A)) * ( d[D(x,y)]/dx )/( 2*sqrt(D(x,y)) )
    #         =  (1/(2*A)) * ( - 4*A*dC(x,y)/dx )/( 2*sqrt(D(x,y)) )
    #         = - ( dC(x,y)/dx )/sqrt(D(x,y)
    #         = - ( Rpole^2*2*x )/sqrt(D(x,y)
    #   df/dy = d[ (-B(y)+sqrt(D(x,y)))/(2*A) ]/dy
    #         = (1/(2*A)) * ( -d[B(y)]/dy + d[sqrt(D(x,y))]/dy )
    #         = (1/(2*A)) * ( -(2*ci*si*(1 - Rpole^2))   + ( (d[D(x,y)]/dy )/( 2*sqrt(D(x,y)) )) )
    #         = (1/(2*A)) * ( -(2*ci*si*(1 - Rpole^2))   + ( ( d[B(y)^2]/dy - d[4*A*C(x,y)]/dy )/( 2*sqrt(D(x,y)) )) )    
    #         = (1/(2*A)) * ( -(2*ci*si*(1 - Rpole^2))   + ( ( 2*B(y)*d[B(y)]/dy - 4*A*d[C(x,y)]/dy  )/( 2*sqrt(D(x,y)) )) )      
    #         = (1/(2*A)) * ( -(2*ci*si*(1 - Rpole^2))   + ( ( 2*y*ci*si*(1 - Rpole^2)*( 2*ci*si*(1 - Rpole^2) ) - 2*A*( 2*y*si^2*(1- Rpole^2) + 2*y*Rpole^2 )  )/sqrt(D(x,y)) ) )        
    #         = (1/(2*A)) * ( -(2*ci*si*(1 - Rpole^2))   + ( ( 4*y*ci^2*si^2*(1 - Rpole^2)^2 - 4*A*y*( si^2*(1- Rpole^2) + Rpole^2 )  )/sqrt(D(x,y)) ) )    
    #         = (1/(2*A)) * ( -(2*ci*si*(1 - Rpole^2))   + 4*y*( ( ci^2*si^2*(1 - Rpole^2)^2 - (1 - si^2*(1-Rpole^2))*( si^2*(1- Rpole^2) + Rpole^2 )  )/sqrt(D(x,y)) ) )  
    #         the term inside the parenthesis simplifies as Rpole^2, so that:
    #         = (1/A) * ( -(ci*si*(1 - Rpole^2))   + 2*y*Rpole^2/sqrt(D(x,y)) ) 
    RpoleReq2 = star_params['RpoleReq']**2.
    ci = np.cos(star_params['istar_rad'])
    si = np.sin(star_params['istar_rad'])
    mRp2 = (1. - RpoleReq2)
    Aquad = 1. - si**2.*mRp2
    Bquad = 2.*y_st_sky_grid*ci*si*mRp2
    Cquad = y_st_sky_grid**2.*si**2.*mRp2 + star_params['RpoleReq']**2.*(x_st_sky_grid**2.+ y_st_sky_grid**2. - 1.) 
    det = Bquad**2.-4.*Aquad*Cquad
    Nx_st_sky_grid2 = ( RpoleReq2*2.*x_st_sky_grid )**2./np.abs(det)
    Ny_st_sky_grid2 = ((1./Aquad)*( -(ci*si*mRp2)   + 2.*y_st_sky_grid*RpoleReq2/np.sqrt(det) ))**2.    
    Nz_st_sky_grid = 1.
    N_grid_star = np.sqrt(Nx_st_sky_grid2 + Ny_st_sky_grid2 + Nz_st_sky_grid*Nz_st_sky_grid )
    mu_grid_star = Nz_st_sky_grid/N_grid_star

    #Calculate gravity-darkening
    if len(gd_band)>0:
        y_pole_star = star_params['RpoleReq']
        gr_R_pole = -G_usi*star_params['Mstar']*Msun/(y_pole_star*Rstar_m)**3.
        g_pole_star = np.abs(y_pole_star*gr_R_pole)  
    
        #Stellar cell temperature
        #    - from Maeder+2009
        temp_grid_star = star_params['Tpole']*(g_grid_star/g_pole_star)**star_params['beta_GD']
    
        #Wavelength table for the band, in A
        nw_band=int((gd_band['wmax']-gd_band['wmin'])/gd_band['dw'])
        dw_band=(gd_band['wmax']-gd_band['wmin'])/nw_band
        wav_band=gd_band['wmin']+0.5*dw_band+dw_band*np.arange(nw_band)            
        
        #Black body flux
        #    - at star surface, but absolute flux does not matter, so we scale by a constant factor to get values around 1
        gd_grid_star = np.sum(planck(wav_band[:,None],temp_grid_star)*dw_band,axis=0)/1e11            

    else:
        gd_grid_star = np.ones(mu_grid_star.shape,dtype=float)

    return gd_grid_star,mu_grid_star
























def get_LD_coeff(transit_prop,iband):
    r"""**Limb-Darkening coefficients**

    Store input limb-darkening coefficients in common structure [LD_u1,LD_u2,LD_u3,LD_u4].

    Args:
        transit_prop (dict) : dictionary containing the planet/spot limb-darkening properties.
        iband (int) : index of the band considered.

    Returns:
        ld_coeff (list) : formatted list of the limb-darkening coefficients used.
    
    """     
    
    LD_mod = transit_prop['LD'][iband]
    #Uniform
    if LD_mod == 'uniform':
        ld_coeff=[np.nan,np.nan,np.nan,np.nan]
    #Linear
    elif LD_mod == 'linear':
        ld_coeff=[transit_prop['LD_u1'][iband],np.nan,np.nan,np.nan]
    #Quadratic, squareroot, logarithmic, power-2, exponential
    elif LD_mod in [ 'quadratic' ,'squareroot','logarithmic', 'power2' ,'exponential']:
        ld_coeff=[transit_prop['LD_u1'][iband],transit_prop['LD_u2'][iband],np.nan,np.nan]
    #Nonlinear
    elif LD_mod == 'nonlinear':               
        ld_coeff=[transit_prop['LD_u1'][iband],transit_prop['LD_u2'][iband],transit_prop['LD_u3'][iband],transit_prop['LD_u4'][iband]]
    #Solar
    elif LD_mod == "Sun":               
        ld_coeff=[transit_prop['LD_u'+str(i)][iband] for i in range(1,7)]
    else:
        stop('Limb-darkening law not supported by ANTARESS')  
    return ld_coeff






def calc_LD(LD_mod,mu,ld_coeff):
    r"""**Limb-Darkening intensity**

    Calculates limb-Darkening value at a given :math:`\mu`

    Args:
        TBD
    
    Returns:
        TBD
    
    """   
    if LD_mod == 'uniform':
        ld_val = 1.
    elif LD_mod == 'linear':
        ld_val = 1. - ld_coeff[0]*(1. -mu)
    elif LD_mod == 'quadratic':
        ld_val = 1. - ld_coeff[0]*(1. -mu) - ld_coeff[1]*np.power(1. -mu,2.)
    elif LD_mod == 'squareroot':
        ld_val = 1. - ld_coeff[0]*(1. -mu) - ld_coeff[1]*(1. - np.sqrt(mu))
    elif LD_mod == 'nonlinear': 
        #Claret+2000
        ld_val = 1. - ld_coeff[0]*(1. -mu**0.5) - ld_coeff[1]*(1. -mu) - ld_coeff[2]*(1. -mu**1.5) - ld_coeff[3]*(1. -mu**2.)
    elif LD_mod == 'power2':        
        ld_val = 1. - ld_coeff[0]*(1. -mu**ld_coeff[1]) 
    elif LD_mod == "Sun":        
        norm_ld = 2.*np.pi*(ld_coeff[0]/2. + ld_coeff[1]/3. - ld_coeff[2]/4. + ld_coeff[3]/5. -ld_coeff[4]/5. +ld_coeff[5]/7.)
        ld_val = (ld_coeff[0] + ld_coeff[1]*mu - ld_coeff[2]*mu**2. +ld_coeff[3]*mu**3. -ld_coeff[4]*mu**4. +ld_coeff[5]*mu**5.)/norm_ld
    return ld_val





def calc_Isurf_grid(iband_list,ngrid_star,system_prop,coord_grid,star_params,Ssub_Sstar,Istar_norm=1.,region = 'star',Ssub_Sstar_ref = None):
    r"""**Stellar intensity grid.**

    Calculates flux and intensity values over the stellar grid, from various contributions.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 

    #Limb-darkening grid and flux emitted by the star
    #    - calculating over the grid as a function of mu, and for each chromatic wavelength given as input  
    #    - mu = cos(theta)), from 1 at the center of the disk to 0 at the limbs), with theta angle between LOS and local normal
    nw = len(iband_list)
    ld_grid_star = np.ones([ngrid_star,nw],dtype=float)
    gd_grid_star = np.ones([ngrid_star,nw],dtype=float)
    mu_grid_star = np.zeros([ngrid_star,nw],dtype=float)
    gd_band={}
    for isubband,iband in enumerate(iband_list):

        #Oblate star with gravity-darkening
        if (star_params['f_GD']>0.):
            if star_params['GD']:gd_band = {'wmin':system_prop['GD_wmin'][iband],'wmax':system_prop['GD_wmax'][iband],'dw':system_prop['GD_dw'][iband]}
            gd_grid_star[:,isubband],mu_grid_star[:,isubband] = calc_GD(coord_grid['x_st'],coord_grid['y_st'],coord_grid['z_st'],star_params,gd_band,coord_grid['x_st_sky'],coord_grid['y_st_sky'])         

        #Spherical star
        else:mu_grid_star[:,isubband] = coord_grid['z_st_sky']

        #Limb-darkening coefficients over the grid
        ld_grid_star[:,isubband]=calc_LD(system_prop['LD'][iband],mu_grid_star[:,isubband],get_LD_coeff(system_prop,iband))

    #Intensity and fluxes from cells over the grid
    #    - the chromatic intensity is normalized so that the mean flux integrated over the defined spectral bands, and summed over the full (normalized) stellar disk, is unity
    # mean(w , sum(cell,F(w,cell))) = 1 
    # mean(w , sum(cell,I(w,cell))*SpSs) = 1 
    # mean(w , sum(cell,I(w,cell))) = 1/SpSs 
    #    - specific intensity at disk center is 1 without LD and GD contributions
    #    - normalized by stellar surface, ie F = I0*Scell/Sstar = 1*dl^2/(pi*Rstar)^2 = (dl/Rstar)^2*(1/pi)
    Isurf_grid_star=ld_grid_star*gd_grid_star
    if Istar_norm==1.:     #calculation for full stellar grid
        Itot_star_chrom = np.sum(Isurf_grid_star,axis=0)
        if system_prop['nw']>1:Istar_norm = np.sum(system_prop['dw'][None,:]*Itot_star_chrom)/np.sum(system_prop['dw'])
        else:Istar_norm = np.mean(Itot_star_chrom)
    #Flux values from stellar cells, normalized by total stellar flux
    if region == 'star':Fsurf_grid_star = Isurf_grid_star / Istar_norm                                  #fcell = I*SpSs/Ftot, where Ftot = sum(I*SpSs) = Itot*SpSs                            
    elif region == 'pl':Fsurf_grid_star = Isurf_grid_star*Ssub_Sstar / (Istar_norm*Ssub_Sstar_ref)      #fcell = I_pl*SpSs_pl/Ftot = I_pl*SpSs_pl/(Itot*SpSs)

    #Total flux over the full star in each band
    Ftot_star = np.sum(Fsurf_grid_star,axis=0)

    return ld_grid_star,gd_grid_star,mu_grid_star,Fsurf_grid_star,Ftot_star,Istar_norm



def calc_st_sky(coord_grid,star_params):
    r"""**Sky-projected stellar grid.**

    Calculates coordinates of stellar cells in the sky-projected star rest frame.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    
    #Distance of subcells to star center (squared)
    coord_grid['r2_st_sky']=coord_grid['x_st_sky']*coord_grid['x_st_sky']+coord_grid['y_st_sky']*coord_grid['y_st_sky']

    #Oblate star
    #    - the condition is that the 2nd order equation yielding zst_sky for a given (xst_sky,yst_sky) has at least one solution
    if star_params['f_GD']>0.:
        coord_grid['z_st_sky'],cond_in_stphot=calc_zLOS_oblate(coord_grid['x_st_sky'],coord_grid['y_st_sky'],star_params['istar_rad'],star_params['RpoleReq'])[1:3]
        if True in cond_in_stphot:
            for key in ['x','y','z','r2']:coord_grid[key+'_st_sky'] = coord_grid[key+'_st_sky'][cond_in_stphot] 
              
    #Spherical star
    #    - mu and z_st_sky are equivalent in this case
    else:
        cond_in_stphot = ( coord_grid['r2_st_sky'] < 1.) 
        if True in cond_in_stphot:
            for key in ['x','y','r2']:coord_grid[key+'_st_sky'] = coord_grid[key+'_st_sky'][cond_in_stphot] 
            coord_grid['z_st_sky'] = np.sqrt(1.- coord_grid['r2_st_sky'])      

    #Frame conversion from the inclined star frame to the 'star' frame 
    #    - positions in star rest frame (in units of stellar radius)
    #    - the (x_st_sky,y_st_sky) coordinates of the planet-occulted region are the same as that of the planet, but the z_st_sky coordinates differ since the region is on the stellar surface
    #      it must thus be calculated directly and not converted from the planet st_sky coordinates
    nsub_star = np.sum(cond_in_stphot)
    if (nsub_star>0):coord_grid['x_st'],coord_grid['y_st'],coord_grid['z_st']=frameconv_skystar_to_star(coord_grid['x_st_sky'],coord_grid['y_st_sky'],coord_grid['z_st_sky'],star_params['istar_rad'])
    
    return nsub_star









def model_star(mode,grid_dic,grid_type,system_prop_in,nsub_Dstar,star_params):
    r"""**Model stellar grid**

    Defines coordinates and intensity values over the stellar grid.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    coord_grid = {}
    
    #Subcell width (in units of Rstar) and surface (in units of Rstar^2 and pi*Rstar^2) 
    d_sub=2./nsub_Dstar      
    Ssub_Sstar=d_sub*d_sub/np.pi    

    #Coordinates of cells discretizing the square enclosing the star
    #    - coordinates are in the sky-projected star frame ('inclined' star frame)
    #    - X axis = projection of stellar equator in the plane of sky
    #      Y axis = projection of the normal to the stellar spin axis
    #      Z axis = line of sight 
    #    - in units of stellar radius
    #    - we first define a regular grid in the sky-projected star frame, rather than in the star frame, because it is the one that is seen during transit and the conversion from star->star inclined would make it irregular in y
    cen_sub=-1.+(np.arange(nsub_Dstar)+0.5)*d_sub            
    xy_st_sky_grid=np.array(list(it_product(cen_sub,cen_sub))) 
    coord_grid['x_st_sky'] = xy_st_sky_grid[:,0] 
    coord_grid['y_st_sky'] = xy_st_sky_grid[:,1]

    #Coordinates in the sky-projected star rest frame
    nsub_star = calc_st_sky(coord_grid,star_params)

    #Storing for model fits
    if mode=='grid':
        grid_dic.update({'Ssub_Sstar' : Ssub_Sstar,'nsub_star' : nsub_star})
        for key in ['x','y','z','r2']: grid_dic[key+'_st_sky'] = coord_grid[key+'_st_sky'] 
        for key in ['x','y','z']: grid_dic[key+'_st'] = coord_grid[key+'_st'] 
        grid_dic['r_proj'] = np.sqrt(coord_grid['r2_st_sky'])  
 
    #Spectral grid
    #    - the grid is used to model disk-integrated profile and thus only need to be chromatic if they are not in CCF format
    #    - all tables nonetheless have the same structure in ncell x nband
    grid_type_eff = ['achrom']
    if ('spec' in grid_type) and ('chrom' in system_prop_in):grid_type_eff+=['chrom']
    for key_type in grid_type_eff:

        #Intensity grid
        ld_grid_star,gd_grid_star,mu_grid_star,Fsurf_grid_star,Ftot_star,Istar_norm = calc_Isurf_grid(range(system_prop_in[key_type]['nw']),nsub_star,system_prop_in[key_type],coord_grid,star_params,Ssub_Sstar)
    
        #Storing for model fits
        if mode=='grid':
            grid_dic.update({'Istar_norm_'+key_type:Istar_norm,
                             'ld_grid_star_'+key_type : ld_grid_star,
                             'gd_grid_star_'+key_type : gd_grid_star,
                             'Fsurf_grid_star_'+key_type : Fsurf_grid_star,
                             'mu_grid_star_'+key_type :  mu_grid_star})

        #Total flux
        #    - correspond to sum(F) = n*dl^2/(pi*Rstar)^2
        elif mode=='Ftot':grid_dic['Ftot_star_'+key_type] = Ftot_star   
        
    if mode=='grid':grid_dic['mu'] = grid_dic['mu_grid_star_achrom'][:,0]
    
    return None


def up_model_star(args,param):
    r"""**Update stellar grid**

    Update coordinates and intensity values over the stellar grid.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    #Update potentially variable stellar properties
    for ideg in range(1,5):
        if ('LD_u'+str(ideg)) in param:args['system_prop']['achrom'].update({'LD_u'+str(ideg):[param['LD_u'+str(ideg)]]}) 
    args['star_params'].update({'f_GD':param['f_GD'],'RpoleReq':1.-param['f_GD'],'beta_GD':param['beta_GD'],'Tpole':param['Tpole'],
                                'istar_rad':np.arccos(param['cos_istar']),'om_eq' : param['veq']/args['star_params']['Rstar_km']})

    #Update stellar grid
    #    - physical sky-projected grid and broadband intensity variations            
    model_star('grid',args['grid_dic'],[args['type']],args['system_prop'],args['grid_dic']['nsub_Dstar'],args['star_params'])

    return None







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from copy import deepcopy
import bindensity as bind


#NB: Kept in an independent file to prevent issues of circular import 
def align_data(data_exp,rout_mode,nord,dim_exp_resamp,resamp_mode,cen_bins_resamp, edge_bins_resamp,rv_shift_cen,spec_dopshift_cen,rv_shift_edge = None ,spec_dopshift_edge = None , nocov = False):
    r"""**Aligning routine.**
    
    Shifts spectral or CCF profiles, and associated tables, with input rv values. 

    Args:
        TBD
    
    Returns:
        TBD
        
    """
    #Shift wavelength tables to a common rest frame
    #    - we apply the general Doppler effect formula (see gen_specdopshift())
    # w_receiver = w_source * (1+ (rv[s/r]/c))
    # w_source = w_receiver / (1+ (rv[s/r]/c))
    #      where rv[s/r] < 0 if the source is moving toward the receiver
    #
    #    - for disk-integrated spectra the source is the star, and the receiver the solar system barycenter (we assume the motion of the Earth with respect to the solar system barycenter has already been corrected)
    #      disk-integrated spectra are aligned in the star rest frame by correcting for the systemic velocity (motion of the stellar system barycenter relative to the solar system barycenter), and
    # the Keplerian motion of the star (motion of the star relative to the stellar system barycenter)
    # w_solbar = w_star * (1+ (rv[star/star bary]/c)) * (1+ (rv[star bary/sun bary]/c))  
    #          = w_star * (1+ (rv_kep/c)) * (1+ (rv_sys/c))   
    # w_star = w_solbar / ((1+ (rv_kep/c)) * (1+ (rv_sys/c))  
    #
    #    - for intrinsic spectra the source is the photosphere, and the receiver the star (this is the frame where they are extracted)
    #      intrinsic spectra are aligned in the photosphere rest frame by correcting for the motion of the stellar surface relative to the star rest velocity
    # w_star = w_surf * (1+ rv[surf/star]/c))
    # w_surf = w_star / (1+ rv[surf/star]/c))
    #
    #    - for out-of-transit profiles or aligned intrinsic profiles used as estimates for a given local profile the source can be seen as the photosphere and the receiver the star (this is the frame where they are defined)
    #      local spectra are shifted back to the star rest frame by re-injecting the motion of the photosphere relative to the star rest velocity
    # w_surf = w_star * (1+ rv[star/surf]/c))    
    # w_star = w_surf / (1+ rv[star/surf]/c))   
    #
    #    - for atmospheric spectra the source is the planet, and the receiver the star (this is the frame where they are extracted)   
    #      atmospheric spectra are aligned in the planet rest frame by correcting for the motion of the planet relative to the star rest velocity
    # w_star = w_pl * (1 + rv[pl/star]/c))  
    # w_pl = w_star / (1 + rv[pl/star]/c)) 
    if ('spec' in rout_mode):
    
        #Achromatic shift
        if (rv_shift_edge is None):
            edge_bins_rest = data_exp['edge_bins']*spec_dopshift_cen
            cen_bins_rest = data_exp['cen_bins']*spec_dopshift_cen                         
         
        #Chromatic shift
        #    - in this case the bin edges and center are shifted using the rv calculated at their exact wavelength, to keep the new bins contiguous
        else:
            edge_bins_rest = data_exp['edge_bins']*spec_dopshift_edge  
            cen_bins_rest = data_exp['cen_bins']*spec_dopshift_cen    

    #Shift velocity tables to chosen rest frame
    #    - for disk-integrated data: RV(M/star) = RV(M/sun) - RV(CDM_star/sun) - RV(star/CDM_star)
    #    - for intrinsic data:       RV(M/region) = RV(M/star) - RV(region/star)
    #    - for master out data:      RV(M/star) = RV(M/region) + RV(region/star)
    #    - for atmospheric data:     RV(M/pl) = RV(M/star) - RV(pl/star)
    elif (rout_mode=='CCF'):
        edge_bins_rest = data_exp['edge_bins'] - rv_shift_cen
        cen_bins_rest = data_exp['cen_bins'] - rv_shift_cen        
        
    #----------------------------------------------------------------

    #Port data from previous processing
    data_align = deepcopy(data_exp) 
    
    #Data is resampled on the common table given as input
    #    - for spectra we neglect the few bins lost by resampling the data (defined on tables shifted to the star rest frame) on the common table defined in the input rest frame
    if cen_bins_resamp is not None:

        #Initialize aligned data
        data_align['edge_bins']=deepcopy(edge_bins_resamp)
        data_align['cen_bins']=deepcopy(cen_bins_resamp) 
        data_align['flux']=np.zeros(dim_exp_resamp, dtype=float)*np.nan
  
        #Aligning each order
        if nocov:
            for iord in range(nord):
                data_align['flux'][iord] = bind.resampling(data_align['edge_bins'][iord], edge_bins_rest[iord], data_exp['flux'][iord], kind=resamp_mode)            
        else:
            data_align['cov']=np.zeros(nord, dtype=object)          
            for iord in range(nord):
                data_align['flux'][iord],data_align['cov'][iord] = bind.resampling(data_align['edge_bins'][iord], edge_bins_rest[iord], data_exp['flux'][iord] , cov = data_exp['cov'][iord], kind=resamp_mode) 
      
        #Processing telluric spectrum
        if ('tell' in data_exp):
            data_align['tell']=np.zeros(dim_exp_resamp, dtype=float)*np.nan
            for iord in range(nord):data_align['tell'][iord] = bind.resampling(data_align['edge_bins'][iord], edge_bins_rest[iord], data_exp['tell'][iord], kind=resamp_mode)            
        
        #Processing calibration profile
        if ('mean_gdet' in data_exp):
            data_align['mean_gdet']=np.zeros(dim_exp_resamp, dtype=float)*np.nan
            for iord in range(nord):data_align['mean_gdet'][iord] = bind.resampling(data_align['edge_bins'][iord], edge_bins_rest[iord], data_exp['mean_gdet'][iord], kind=resamp_mode)             
            
        #Defined bins
        data_align['cond_def'] = ~np.isnan(data_align['flux'])
 
    #Data remain defined on independent tables for each exposure
    #    - we do not resample the data tables of each exposure and keep them defined on their shifted spectral tables
    else:      
        data_align['cen_bins'] = cen_bins_rest   
        data_align['edge_bins'] = edge_bins_rest                

    return data_align




